# -------------------------------------------
#
# GenSol Package
#
# Python tool for simulation of abstract photonic circuits
#
# @author: Marco Passoni
#
# ------------------------------------------
from typing import Any, List, Dict, Tuple, Callable

import numpy as np
from copy import copy
from copy import deepcopy
from lekkersim.structure import Structure
from lekkersim import sol_list
from lekkersim import logger
from lekkersim.utils import map_args
from lekkersim.model import SolvedModel


class Solver:
    """Class Solver
    This class defines the simulations. It contains all the structures of the optical componets, and has the methods for running the simulation and accessing the results.

    Args:
        structures (list) : list of structures in the solver. Default is None (empty list)
        connections (dict) : dictionary of tuples (structure (Structure), pin (str)) containing connections {(structure1,pin1):(structure2,pin2)}. Default is None (empty dictionary)
        param_dic (dict) :  dictionary of parameters {param_name (str) : param_value (usually float)}. Default is None (empty dictionary). The provided values are assume as the default parameters
        param_mapping (dict): dict for re-definition of the parameters. Useful to include some kind of physic level knowledge in the model. The dictionary is build as:
            >>> {'old_name' : (func, {'new_name' : new_default_vaule})}.
            where func is a function whose parameter are provided by the second dictionary. When simulating, the values of the parameters in the 'new_name' dictionary are passed to func and the retunrded values assigned to 'old_name'.

    """

    structures: List[Structure]

    space = ""
    depth = 0

    def __init__(
        self,
        name: str = None,
        structures: List[Structure] = None,
        connections: Dict[Tuple[Structure, str], Tuple[Structure, str]] = None,
        pin_mapping: Dict[str, Tuple[Structure, str]] = None,
        param_dic: Dict[str, Any] = None,
        param_mapping: Dict[str, str] = None,
    ) -> None:
        """Creator"""
        self.structures = structures if structures is not None else []
        self.connections = connections if connections is not None else {}
        self.connections_list = []
        self.param_dic = param_dic if param_dic is not None else {}
        self.pin_mapping = pin_mapping if pin_mapping is not None else {}
        self.default_params = {"wl": None}
        self.default_params.update(self.param_dic)
        self.param_mapping = {} if param_mapping is None else param_mapping
        self.monitor_st = {}
        for pin1, pin2 in self.connections.items():
            self.connections_list.append(pin1)
            self.connections_list.append(pin2)
            pin1[0].add_conn(pin1[1], *pin2)
            pin2[0].add_conn(pin2[1], *pin1)
        if len(set(self.connections_list)) != len(self.connections_list):
            raise ValueError("Same pin connected multiple time")
        self.free_pins = []
        for st in self.structures:
            for pin in st.pin_list:
                self.free_pins.append(pin)
        for pin in self.connections_list:
            self.free_pins.remove(pin)
        self.name = name

    def __enter__(self):
        """Make the Solver the active solver

        Usage:
            >>> with Solver() as MySol:
            >>>     stuff


        Until the with statement is closed, every change (for example, from put methods) will be applied to MySol
        """
        sol_list.append(self)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Delete the solver from the list of the active ones."""
        sol_list.pop()

    def __str__(self):
        """Formatter for printing"""
        if self.name is None:
            return f"Solver object (id={id(self)})"
        else:
            return f"Solver of {self.name} (id={id(self)})"

    def is_empty(self) -> bool:
        """Checks if solver is empy

        Returns:
            Bool: False is model has no pins, True otherwise

        """
        if len(self.structures) > 0:
            return False
        else:
            return True

    def add_structure(self, structure: Structure) -> None:
        """Add a structure to the solver

        Args:
            structure (Structure) : structure to be added

        Returns:
            None
        """
        if structure not in self.structures:
            self.structures.append(structure)
            for pin in structure.pin_list:
                self.free_pins.append(pin)
        else:
            raise ValueError("Structure already present")

        inv_mapping = {
            old_name: new_name for new_name, old_name in structure.param_mapping.items()
        }
        default_dic = {}
        if structure.model is not None:
            default_params = structure.model.default_params
        elif structure.solver is not None:
            default_params = structure.solver.default_params
        else:
            default_params = {}

        for key, value in default_params.items():
            if key in ["R", "w", "pol"]:
                continue
            if key in inv_mapping:
                default_dic[inv_mapping[key]] = value
            else:
                default_dic[key] = value
        self.default_params.update(default_dic)

    def cut_structure(self, structure: Structure) -> None:
        """Remove structure from solver, cutting all the connections to other structures (pins in connected structure are not removed, but freed again)

        Args:
            structure (Structure) : structure to be removed

        Returns:
            None
        """
        if structure not in self.structures:
            raise Exception("Structure {structure} is not in solver {self}")
        self.structures.remove(structure)
        for st in structure.connected_to:
            st.cut_connections(structure)
        copy_dic = copy(self.connections)
        for (st1, pin1), (st2, pin2) in copy_dic.items():
            if st1 is structure or st2 is structure:
                self.connections.pop((st1, pin1))
                self.free_pins.append((st2, pin2))
                self.free_pins.append((st1, pin1))
                self.connections_list.remove((st2, pin2))
                self.connections_list.remove((st1, pin1))
        copy_dic = copy(self.free_pins)
        for st, pin in copy_dic:
            if st is structure:
                self.free_pins.remove((st, pin))
        copy_dic = copy(self.pin_mapping)
        for pinname, (st, pin) in copy_dic.items():
            if st is structure:
                self.pin_mapping.pop(pinname)

    def remove_structure(self, structure: Structure) -> None:
        """Remove structure from solver, also removing all the connections to other structures

        Args:
            structure (Structure) : structure to be removed

        Returns:
            None
        """
        if structure not in self.structures:
            raise Exception("Structure {structure} is not in solver {self}")
        self.structures.remove(structure)
        for st in structure.connected_to:
            st.remove_connections(structure)
        copy_dic = copy(self.connections)
        for (st1, pin1), (st2, pin2) in copy_dic.items():
            if st1 is structure:
                self.connections.pop((st1, pin1))
            if st2 is structure:
                self.connections.pop((st1, pin1))
        copy_dic = copy(self.free_pins)
        for st, pin in copy_dic:
            if st is structure:
                self.free_pins.remove((st, pin))
        copy_dic = copy(self.pin_mapping)
        for pinname, (st, pin) in copy_dic.items():
            if st is structure:
                self.pin_mapping.pop(pinname)

    def monitor_structure(
        self, structure: Structure = None, name: str = "Monitor"
    ) -> None:
        """Add structure to the ones to be monitored for internal modes

        Args:
            structure (Structure): structure to be added
            name (string): name to the associated with this monitor. Default is 'Monitor'
        """
        if structure not in self.structures:
            raise ("Error: structure {structure} not in {self}")
        self.monitor_st[structure] = name

    def connect(
        self, structure1: Structure, pin1: str, structure2: Structure, pin2: str
    ) -> None:
        """Connect two different structures in the solver by the specified pins

        Args:
            structure1 (Structure) : first structure
            pin1 (str) : pin of first structure
            structure2 (Structure) : second structure
            pin2 (str) : pin of second structure

        Returns:
            None
        """
        if (structure1, pin1) in self.connections_list:
            if (structure1, pin1) in self.connections and self.connections[
                (structure1, pin1)
            ] == (structure2, pin2):
                return
            if (structure2, pin2) in self.connections and self.connections[
                (structure2, pin2)
            ] == (structure1, pin1):
                return
            raise ValueError("Pin already connected")
        if (structure2, pin2) in self.connections_list:
            raise ValueError("Pin already connected")
        self.connections_list.append((structure1, pin1))
        self.connections_list.append((structure2, pin2))
        self.connections[(structure1, pin1)] = (structure2, pin2)
        self.free_pins.remove((structure1, pin1))
        self.free_pins.remove((structure2, pin2))
        structure1.add_conn(pin1, structure2, pin2)
        structure2.add_conn(pin2, structure1, pin1)

    def connect_all(
        self, structure1: Structure, pin1: str, structure2: Structure, pin2: str
    ) -> None:
        """Connect the two structures using all the pins with the matching basename

        Args:
            structure1 (Structure): first structure
            pin1 (str): basename of pin in structure1
            structure2 (Structure): second structure
            pin2 (str): basename of pin in structure2

        Returns:
            None
        """
        modes1 = set(structure1.get_pin_modenames(pin1))
        modes2 = set(structure2.get_pin_modenames(pin2))
        if modes1 != modes2:
            logger.error(
                f"{pin1} in {structure1} and {pin2} in {structure2} and have differents modes: only matching modes connected"
            )
        modes = modes1.intersection(modes2)
        for m in modes:
            p1 = pin1 if m == "" else "_".join([pin1, m])
            p2 = pin2 if m == "" else "_".join([pin2, m])
            self.connect(structure1, p1, structure2, p2)

    def show_free_pins(self) -> None:
        """Print all pins of the structure in the solver whcih are not connected. If a pin mapping exists, it is also reported"""
        print("Free pins of solver: %50s)" % (self))
        for st, pin in self.free_pins:
            try:
                pinname = list(self.pin_mapping.keys())[
                    list(self.pin_mapping.values()).index((st, pin))
                ]
                print("  (%50s, %5s) --> %5s" % (st, pin, pinname))
            except ValueError:
                print("  (%50s, %5s)" % (st, pin))

    def show_structures(self) -> None:
        """Print all structures in the solver"""
        print("Structures and pins of solver: %50s)" % (self))
        for st in self.structures:
            print("  %50s" % (st))

    def show_connections(self) -> None:
        """Print all connected pins in the solver"""
        print("Connection of solver: %50s)" % (self))
        for c1, c2 in self.connections.items():
            print("  (%50s, %5s) <--> (%50s, %5s)" % (c1 + c2))

    def show_pin_mapping(self) -> None:
        """If a pin mapping is defined, print only mapped pins"""
        try:
            for pinname, (st, pin) in self.pin_mapping.items():
                print("  %5s --> (%50s, %5s)" % (pinname, st, pin))
        except AttributeError:
            print("  No mapping defined")

    def map_pins(self, pin_mapping: Dict[str, Tuple[Structure, str]]):
        """Add mapping of pins

        Args:
            pin_mapping (dict): dictionary of pin mapping in the form {pin_name (str) : (structure (Structure), pin (str) )}

        Returns:
            None
        """
        self.pin_mapping.update(pin_mapping)

    def solve(self, **kwargs) -> SolvedModel:
        """Calculates the scattering matrix of the solver

        Args:
            kwargs (dict) : paramters in the form param_name=param_value

        Returs:
            SolvedModel : model containing the scattering matrix
        """
        logger.debug(f"Solving {self}")
        if len(self.free_pins) > len(self.pin_mapping):
            logger.warning(
                f"{self}:Running solve without complete mapping: some pins will not be not accessible"
            )
        func = None
        monitor_mapping = None
        self.update_params(kwargs)
        ns = 1
        for par, value in self.param_dic.items():
            self.param_dic[par] = np.reshape(value, -1)
            if len(self.param_dic[par]) == 1:
                continue
            if ns == 1:
                ns = len(self.param_dic[par])
            elif ns != len(self.param_dic[par]):
                raise ValueError("Calling solve with parameters of different shapes")

        for par, value in self.param_dic.items():
            self.param_dic[par] = (
                np.array([value[0] for i in range(ns)]) if len(value) == 1 else value
            )
        for st in self.structures:
            st.update_params(self.param_dic)
        st_list = [st for st in self.structures if st not in self.monitor_st]
        if len(st_list) == 1:
            st_list[0].createS()
        while len(st_list) != 1:
            st_list_pins = [st.pin_count for st in st_list]
            st_list = [st_list[i] for i in np.argsort(st_list_pins)]
            source_st = st_list[0].gone_to
            connected_to = [
                st_list[0].connected_to[i]
                for i in np.argsort(
                    [_.gone_to.pin_count for _ in st_list[0].connected_to]
                )
            ]
            for st in connected_to + st_list[1:]:
                if st.gone_to in st_list:
                    tar_st = st.gone_to
                    break
            new_st = source_st.join(tar_st)
            st_list.remove(source_st)
            st_list.remove(tar_st)
            st_list.append(new_st)
        self.main = st_list[0]

        if self.monitor_st != {}:
            st_list = list(self.monitor_st.keys())
            if len(st_list) == 1:
                st_list[0].createS()
            while len(st_list) != 1:
                source_st = st_list[0].gone_to
                for st in st_list[0].connected_to + st_list[1:]:
                    if st.gone_to in st_list:
                        tar_st = st.gone_to
                        break
                new_st = source_st.join(tar_st)
                st_list.remove(source_st)
                st_list.remove(tar_st)
                st_list.append(new_st)
            self.monitor = st_list[0]
            self.total = self.main.join(self.monitor)
            func, pins = self.main.intermediate(self.monitor, self.pin_mapping)
            monitor_mapping = {
                f"{self.monitor_st[st]}_{pin}": i for (st, pin), i in pins.items()
            }

        else:
            self.total = self.main

        for st in self.structures:
            st.reset()
        mod = self.total.get_model(self.pin_mapping, name=self.name)
        mod.solved_params = deepcopy(self.param_dic)
        if func is not None:
            mod.set_intermediate(func, monitor_mapping)
        self.param_dic = {}
        return mod

    def set_param(self, name: str, value: Any = None) -> None:
        """Set a value for one parameter. This is assued as the new default

        Args:
            name (str) : name of the parameter
            value (usually float) : value of the parameter

        Returns:
            None
        """
        self.default_params.update({name: value})

    def put(
        self,
        source_pin: str = None,
        target_pin: Tuple[Structure, str] = None,
        param_mapping: Dict[str, str] = None,
    ) -> Structure:
        """Function for putting a Solver in another Solver object, and eventually specify connections

        This function creates a Structure object for the Solver and place it in the current active Solver
        If both pins and pint are provided, the connection also is made.

        Args:
            source_pin (str): pin of model to be connected
            target_pin (tuple): tuple (structure (Structure) , pin (str)) existing structure and pin to which to connect pins of model
            param_mapping (dict): dictionary of {oldname (str) : newname (str)} containing the mapping of the names of the parameters

        Returns:
            Structure: the Structure instance created from the Solver
        """
        if param_mapping is None:
            param_mapping = {}
        # ST=Structure(solver=deepcopy(self),param_mapping=param_mapping)
        ST = Structure(solver=self, param_mapping=param_mapping)
        sol_list[-1].add_structure(ST)
        if (source_pin is not None) and (target_pin is not None):
            sol_list[-1].connect(ST, source_pin, target_pin[0], target_pin[1])

        #        default_dic={}
        #        for key, value in self.default_params.items():
        #            if key in ['R','w','wl']: continue
        #            if key in param_mapping:
        #                default_dic[param_mapping[key]] = value
        #            else:
        #                default_dic[key] = value
        #        sol_list[-1].default_params.update(default_dic)

        return ST

    def shallow_copy(self):
        """Build a shallow copy of the solver

        This function build a copy of the original solver. Structure objject of the new solver are copies of the old ones. If a structure reference to a model or a solver, the copied structure will still refer to the old model or solver.

        Returs:
            Solver: Shallow copy of the solver
        """
        copy_dic = {}
        new_structures = []
        for st in self.structures:
            if st.model is not None:
                new_st = Structure(
                    model=st.model,
                    param_mapping={val: key for key, val in st.param_mapping.items()},
                )
            elif st.solver is not None:
                new_st = Structure(
                    solver=st.solver,
                    param_mapping={val: key for key, val in st.param_mapping.items()},
                )
            else:
                new_st = deepcopy(st)
            new_structures.append(new_st)
            copy_dic[st] = new_st
        new_connections = {}
        for (st1, pin1), (st2, pin2) in self.connections.items():
            new_connections[(copy_dic[st1], pin1)] = (copy_dic[st2], pin2)
        sol = Solver(
            structures=new_structures,
            connections=new_connections,
            param_dic=self.default_params.copy(),
        )
        pin_mapping = {
            name: (copy_dic[st], pin) for name, (st, pin) in self.pin_mapping.items()
        }
        sol.map_pins(pin_mapping)
        sol.param_mapping = self.param_mapping.copy()
        return sol

    def flatten_top_level(self) -> bool:
        """Flatten top level of a solver

        Reduce the depth of the solver of one level, starting from the top.
        If depth is already 1, nothing is done and the function returns False

        Returns:
            bool: True if reduction is done, False if depth is already 1
        """

        solvers = [
            structure for structure in self.structures if structure.solver is not None
        ]
        for st in solvers:
            st.solver = st.solver.shallow_copy()
        old_conn = copy(self.connections)
        old_mapping = copy(self.pin_mapping)
        if solvers == []:
            return False

        for st in solvers:
            self.cut_structure(st)
            if st.solver.param_mapping != {}:
                logger.error(
                    "During flattening: solver {st.solver} has custom parametes. Simulation results will not be correct."
                )
            for lower_st in st.solver.structures:
                self.add_structure(lower_st)
            for pin1, pin2 in st.solver.connections.items():
                self.connect(*pin1, *pin2)
        for tup1, tup2 in old_conn.items():
            if tup1[0] in solvers:
                if tup2[0] in solvers:
                    self.connect(
                        *tup1[0].solver.pin_mapping[tup1[1]],
                        *tup2[0].solver.pin_mapping[tup2[1]],
                    )
                else:
                    self.connect(*tup1[0].solver.pin_mapping[tup1[1]], *tup2)
            elif tup2[0] in solvers:
                self.connect(*tup1[0], *tup2[0].solver.pin_mapping[tup2[1]])
        new_mapping = {}
        for pin, tup in old_mapping.items():
            if tup[0] in solvers:
                new_mapping[pin] = tup[0].solver.pin_mapping[tup[1]]
        self.map_pins(new_mapping)

        for st in solvers:
            up_dic = {}
            for lower_st in st.solver.structures:
                up_dic[lower_st] = {}
                for top, middle in st.param_mapping.items():
                    if middle in lower_st.param_mapping:
                        bottom = lower_st.param_mapping.pop(middle)
                        up_dic[lower_st][top] = bottom
                    elif top not in lower_st.param_mapping:
                        up_dic[lower_st][top] = middle
            for lower_st in st.solver.structures:
                lower_st.param_mapping.update(up_dic[lower_st])

        return True

    def flatten(self) -> None:
        """Collapse the hyerarchycal structure of the solver in only one level.

        Returns:
            None
        """
        dec = True
        while dec:
            dec = self.flatten_top_level()
        return None

    def _inspect(self, max_depth: int = None) -> None:
        """Recursive function for printing one step of the solver hierarchy

        Args:
            max_depth (int, optional): Maximum depth for printing the circuit. Default is None (full circuit is printed)

        Returns:
            None
        """
        for s in self.structures:
            if max_depth is None or self.__class__.depth < max_depth:
                print(f"{self.space}  {s}")
            if s.solver is not None:
                self.__class__.space = self.__class__.space + "  "
                self.__class__.depth += 1
                s.solver._inspect(max_depth=max_depth)
                self.__class__.space = self.__class__.space[:-2]
                self.__class__.depth -= 1

    def inspect(self, max_depth: int = None) -> None:
        """Print the full hierarchy of the solver

        Args:
            max_depth (int, optional): Maximum depth for printing the circuit. Default is None (full circuit is printed)

        Returns:
            None
        """
        print(f"{self.space}{self}")
        self._inspect(max_depth=max_depth)

    def show_default_params(self) -> None:
        """Print the names of all the top-level parameters and corresponding default value"""
        print(f"Default params of {self}:")
        for name, par in self.default_params.items():
            print(f"  {name:10}: {par}")

    def maps_all_pins(self) -> None:
        """Function for automatically map all pins.

        It scans the unmapped pins and raise them at top level wiht the same name. If one or more pins have the same name, it fails.
        It ignores any pin already mapped by the user.
        """
        for st, pin in self.free_pins:
            if (st, pin) in self.pin_mapping.values():
                continue
            if pin in self.pin_mapping:
                raise Exception("Pins double naming present, cannot map authomatically")
            self.pin_mapping[pin] = (st, pin)

    def update_params(self, update_dic: Dict[str, Any]) -> None:
        """Update the parameters of solver, setting defaults when value is not provides. It takes care of any parameter added with add_param.

        Args:
            update_dic (dict) : dictionary of parameters in the from {param_name (str) : param_value (usually float)}

        Returns:
            None
        """
        # self.param_dic.update(self.default_params)
        # self.param_dic.update(update_dic)
        start_dic = {}
        for name, (func, args) in self.param_mapping.items():
            new_args = {key: value for key, value in args.items()}
            for key, value in new_args.items():
                if key in update_dic:
                    new_args[key] = update_dic[key]
                new_value = func(**new_args)
            start_dic.update({name: new_value})
        self.param_dic.update(self.default_params)
        self.param_dic.update(update_dic)
        self.param_dic.update(start_dic)

    def add_param(
        self, old_name: str, func: Callable, default: Dict[str, Any] = None
    ) -> None:
        """Define a paramter of the solver in term of new paramter(s)

        Args:
            old_name (str) : name of the old parameter to set
            func (function): function linking old parameter to new parameter(s)
            default (dic)  : default value(s) of the new parameter(s).
                Can be None. In this case, introspection is used to try fo find the new parameter(s) and default from func. An error is raised if not possible.
        """
        if default is None:
            var_names = func.__code__.co_varnames
            var_def = func.__defaults__
            if len(var_names) != len(var_def):
                raise ValueError("Not all default provided")
            default = {key: value for key, value in zip(var_names, var_def)}
        up = {old_name: (func, default)}
        self.param_mapping.update(up)
        self.default_params.pop(old_name)
        self.default_params.update(default)

    def prune(self) -> bool:
        """Remove dead branch in the solver hierarchy (the ones ending with an empy solver)

        Returns:
            bool: True if Solver is empty
        """
        # print(f'Entered in {solver}')
        not_empty = []
        copy_list = copy(self.structures)
        for st in copy_list:
            if st.model is not None:
                if st.model.is_empty():
                    self.remove_structure(st)
                else:
                    not_empty.append(st)
                continue
            if st.solver is not None:
                if st.solver.prune():
                    self.remove_structure(st)
                else:
                    not_empty.append(st)
        return len(not_empty) == 0

    def split(self) -> List:
        """Identify sub-circuits inside the solver and splits them into multiple solvers

        Returns:
            list: list of the solvers containing the sub_circuits
        """
        sets = []
        for st in self.structures:
            connected_sets = []
            for _set in sets:
                for target in st.connected_to:
                    if target in _set:
                        _set.add(st)
                        connected_sets.append(_set)
            for _set in connected_sets:
                sets.remove(_set)
            if len(connected_sets) == 0:
                sets.append(set([st] + st.connected_to))
            else:
                sets.append(set().union(*connected_sets))
        solvers = []
        for i, _set in enumerate(sets):
            connections = {
                t1: t2 for t1, t2 in self.connections.items() if t1[0] in _set
            }
            mapping = {
                name: pin for name, pin in self.pin_mapping.items() if pin[0] in _set
            }
            solvers.append(
                Solver(
                    name=f"{self.name}_{i}",
                    structures=list(_set),
                    connections=connections,
                    pin_mapping=mapping,
                    param_mapping=self.param_mapping,
                    param_dic=self.param_dic,
                )
            )
        return solvers


class Pin:
    """Helper class for more user friendly pin mapping (same as Nazca sintax"""

    def __init__(self, name: str, pin: Tuple[Structure, str] = None):
        """

        Args:
            name (str) : name of the pin
            pin (tuple) : tuple of (structure (Structure), pin (str)) containing the data to the pin to be mapped
        """
        self.name = name
        self.pin = None

    def put(self, pin: Tuple[Structure, str] = None):
        """Maps the pins in the tuple to self.name

        Args:
            pin (tuple) : tuple of (structure (Structure), pin (str)) containing the data to the pin to be mapped
        """
        if pin is not None:
            self.pin = pin

        if self.pin is not None:
            sol_list[-1].map_pins({self.name: pin})


sol_list.append(Solver())


def putpin(name: str, tup: Tuple[Structure, str]) -> None:
    """Maps a pin of the current active solver

    Args:
        name (str) : name of the new pin
        tup (tuple) : tuple of (structure (Structure), pin (str)) containing the data to the pin to be mapped
    """
    sol_list[-1].map_pins({name: tup})


def connect(tup1: Tuple[Structure, str], tup2: Tuple[Structure, str]) -> None:
    """Connect two structures in the active Solver

    Args:
        tup1 (tuple) : tuple of (structure (Structure), pin (str)) containing the data of the first pin
        tup1 (tuple) : tuple of (structure (Structure), pin (str)) containing the data of the second pin
    """
    sol_list[-1].connect(tup1[0], tup1[1], tup2[0], tup2[1])


def connect_all(
    structure1: Structure, pin1: str, structure2: Structure, pin2: str
) -> None:
    """Connect in the active solver the two structures using all the pins with the matching basename

    Args:
        structure1 (Structure): first structure
        pin1 (str): basename of pin in structure1
        structure2 (Structure): second structure
        pin2 (str): basename of pin in structure2

    Returns:
        None
    """
    sol_list[-1].connect_all(structure1, pin1, structure2, pin2)


def add_param(old_name: str, func: Callable, default: Dict[str, Any] = None) -> None:
    """Define a paramter of the active solver in term of new paramter(s)

    Args:
        old_name (str) : name of the old parameter to set
        func (function): function linking old parameter to new parameter(s)
        default (dic)  : default value(s) of the new parameter(s).
            Can be None. In this case, introspection is used to try fo find the new parameter(s) and default from func. An error is raised if not possible.
    """
    sol_list[-1].add_param(old_name, func, default=default)


def set_default_params(dic: Dict[str, Any]) -> None:
    """Set default parameters for the solver

    The provided dict will oervwrite the default parameters. All pre-existing parameters will be deleted

     Args:
         dic (dict): dictionary of the default parameters {param_name (str) : default_value (usually float)}
    """
    sol_list[-1].default_params = dic


def update_default_params(dic: Dict[str, Any]) -> None:
    """Update default parameters for the solver

    The provided dict will upadte the default parametes. Not included pre-existing parmeters will be kept.

    Args:
        dic (dict): dictionary of the default parameters {param_name (str) : default_value (usually float)}
    """
    sol_list[-1].default_params.update(dic)


def raise_pins() -> None:
    """Raise all pins in the solver. It reuiqres unique pin naming, otherwaise an error is raised

    It scans the unmapped pins and raise them at top level wiht the same name. If one or more pins have the same name, it fails.
    It ignores any pin already mapped by the user.
    """
    sol_list[-1].maps_all_pins()


def solve(**kwargs) -> SolvedModel:
    """Solve active solver and returns the model

    Args:
        **kwargs : parameters for the simulation

    Returns:
        Model: Model of the active solver.
    """
    return sol_list[-1].solve(**kwargs)


def add_structure_to_monitors(structure: Structure, name: str = "Monitor") -> None:
    """Add structure to the ones to be monitored for internal modes. It effects the active solver

    Args:
        structure (Structure): structure to be added
        name (string): name to the associated with this monitor. Default is 'Monitor'
    """
    sol_list[-1].monitor_structure(structure=structure, name=name)



"""Code for Nazca integrazion

IN order to work it require a custom version of Nazca that is not released. Do not use.
"""


from __future__ import annotations
from typing import Any, Callable, Union

import numpy as np
from copy import copy
from copy import deepcopy
from itertools import tee
import functools

import nazca as nd

from .model import Model
from .structure import Structure
from .log import logger
from .utils import ProtectedPartial
from .sol import Solver


class Model_from_NazcaCM(Model):
    """Class for model from a nazca cell with compact models"""

    def __init__(
        self,
        cell: nd.Cell,
        ampl_model: str = None,
        loss_model: str = None,
        optlength_model: str = None,
        allowed: dict[str, dict[str, Any]] = None,
    ) -> None:
        """Initialize the model

        This class is for building scattering matrices from the connectivity infomation built in nazca.

        In the definition of the scattering matrices, tree models can be provided.
            - ampl_model: This model returns directly the amplitude complex coefficient A.
            - loss_model: This model returns the loss used for the calculation of the modulus of the amplitude (in dB):
                abs(A)^2 = 10^loss/10.0
            - optlength_model: This model returns the optical length used for the calculation of the phase of the
                amplitude: angle(A) = 2.0*pi/wl*optlenght. It assumes the same unit as the wavelegnth.

        The priority of usage withing the models is the following.
            1. ampl: is the model for amplitude is found, the others are ignored. Amplitude between the same pin
                (reflection) is assumed 0 if not otherwaise specified.
            2. loss and optlength: if ampl  is not found, a model will be build using the available information
                between loss and optlength. If either one is missing, it will be assumed as 0
                (i.d. no loss means abs(A)=1, no phase means A purely real), but a warning will be raised.

        Args:
            cell (Nazca Cell): it expects a Nazca cell with some compact model defined.
            ampl_model (str): model to the used to diectly get the scattering matrix amplitudes
            loss_model (str): model to be used to estimate the loss
            optlength_model (str): model to be used for the optical length (phase)
            allowed (dict): mapping {Mode:extra}. The allowed mode in the cell and the extra information to pass to
                the compact model to build the optical length.
        """
        self.cell = cell
        self.ampl_model = ampl_model
        self.loss_model = loss_model
        self.optlength_model = optlength_model
        self.allowed = allowed

        self.name = getattr(cell, "cell_name", cell.name)
        self.pin_dic = {}
        self.param_dic = {}
        self.default_params = {}

        if self._search_amplitude_model():
            return
        self._search_length_and_loss_model()

    def _search_amplitude_model(self):
        """Creates scattering matrix compact model from amplitude model"""
        opt_conn = {}
        # Checking for ampl model
        for name, pin in self.cell.pin.items():
            for mode, extra_in in self.allowed.items():
                opt = list(pin.path_nb_iter(self.ampl_model, extra=extra_in))
                if len(opt) != 0:
                    opt_conn[(pin, mode)] = opt
        if len(opt_conn) == 0:
            return False

        n = 0
        self.CM = {}
        pins = set([pin for pin, mode in opt_conn])
        for pi in pins:
            for mi in self.allowed:
                pin_in = pi.basename if mi == "" else "_".join([pi.basename, mi])
                self.pin_dic[pin_in] = n
                n += 1
                for po in pins:
                    for mo in self.allowed:
                        pin_out = (
                            po.basename if mo == "" else "_".join([po.basename, mo])
                        )
                        tup = tup = (pin_in, pin_out)
                        self.CM[tup] = 0.0

        logger.debug(f"Model for {self.name}: using amplitude model {self.ampl_model}")
        self.N = len(self.pin_dic)
        for (pin, mode), conn in opt_conn.items():
            for stuff in conn:
                target = stuff[0]
                CM = stuff[1]
                extra_out = stuff[5]
                pin_in = pin.basename if mode == "" else "_".join([pin.basename, mode])
                try:
                    modeo = (
                        mode
                        if extra_out is None
                        else list(self.allowed.keys())[
                            list(self.allowed.values()).index(extra_out)
                        ]
                    )
                except ValueError:
                    logger.error(
                        f"Model for {self.name}: mode {extra_out} is not in allowed {self.allowed}: ignored"
                    )
                    continue
                pin_out = (
                    target.basename
                    if modeo == ""
                    else "_".join([target.basename, modeo])
                )
                tup = (pin_in, pin_out)
                self.CM[tup] = (
                    self.wraps(ProtectedPartial(CM, **self.allowed[mode]))
                    if callable(CM)
                    else CM
                )
        return True

    def _search_length_and_loss_model(self):
        """Creates scattering matrix compact model from optical length and loss models"""
        opt_conn = {}
        n = 0
        for name, pin in self.cell.pin.items():
            for mode, extra_in in self.allowed.items():
                opt = {
                    (x[0], str(x[-1])): x[1:]
                    for x in pin.path_nb_iter(self.optlength_model, extra=extra_in)
                }
                lss = {
                    (x[0], str(x[-1])): x[1:]
                    for x in pin.path_nb_iter(self.loss_model, extra=extra_in)
                }
                for target in set(opt.keys()).union(set(lss.keys())):
                    if (pin, mode) not in opt_conn:
                        opt_conn[(pin, mode)] = {}
                    tup1 = (
                        opt[target]
                        if target in opt
                        else (0.0, None, None, None, self.allowed[mode])
                    )
                    tup2 = (
                        lss[target]
                        if target in lss
                        else (0.0, None, None, None, self.allowed[mode])
                    )
                    opt_conn[(pin, mode)][target] = tup1 + tup2
        logger.debug(
            f"Model for {self.name}: using optical length model {self.optlength_model} and loss model {self.loss_model}"
        )
        pins = set([pin for pin, mode in opt_conn])
        self.CM = {}
        for pi in pins:
            for mi in self.allowed:
                pin_in = pi.basename if mi == "" else "_".join([pi.basename, mi])
                self.pin_dic[pin_in] = n
                n += 1
                for po in pins:
                    for mo in self.allowed:
                        pin_out = (
                            po.basename if mo == "" else "_".join([po.basename, mo])
                        )
                        tup = tup = (pin_in, pin_out)
                        self.CM[tup] = 0.0

        self.N = len(self.pin_dic)

        for (pin, mode), conn in opt_conn.items():
            for (target, extra_target), stuff in conn.items():
                OM = stuff[0]
                extra_om = stuff[4]
                AM = stuff[5]
                extra_am = stuff[9]
                if extra_om != extra_am:
                    if extra_om == self.allowed[mode]:
                        extra_om = extra_am

                pin_in = pin.basename if mode == "" else "_".join([pin.basename, mode])
                try:
                    modeo = (
                        mode
                        if extra_om is None
                        else list(self.allowed.keys())[
                            list(self.allowed.values()).index(extra_om)
                        ]
                    )
                except ValueError:
                    logger.error(
                        f"Model for {self.name}: mode {extra_om} is not in allowed {self.allowed}: ignored"
                    )
                    continue
                pin_out = (
                    target.basename
                    if modeo == ""
                    else "_".join([target.basename, modeo])
                )
                tup = (pin_in, pin_out)
                OMt = ProtectedPartial(OM, **self.allowed[mode]) if callable(OM) else OM
                AMt = ProtectedPartial(AM, **self.allowed[mode]) if callable(AM) else AM
                self.CM[tup] = self.__class__.generator(OMt, AMt)

    @staticmethod
    def call_partial(func: functools.partial, *args, **kwargs) -> Any:
        """Evaluate a partial function with args ad kwargs provided and the args and kwargs saved at partial creation

        If any of the keywork arguments is not supported by the original function (the one used to create the partial),
        the keyword is ignored and a warning is logged.

        Args:
            func (functools.partial): partial function to be used
            args: positional arguments
            kwargs: keyword arguments

        Returns:
            Any: whatever the function returns after the unaccepted keyworkd arguments are removed.
        """
        while True:
            to_remove = getattr(func, "to_remove", [])
            for key in to_remove:
                func.keywords.pop(key, None)
                kwargs.pop(key, None)
            try:
                return func(*args, **kwargs)
            except TypeError as e:
                if "got an unexpected keyword argument" not in e.args[0]:
                    raise e
                else:
                    var = e.args[0].split("'")[-2]
                try:
                    func.to_remove.append(var)
                except AttributeError:
                    func.to_remove = [var]
                obj = func
                while True:
                    try:
                        name = obj.__name__
                        break
                    except AttributeError:
                        obj = obj.func
                code = obj.__code__
                logger.warning(
                    f'Function "{name}" in  file "{code.co_filename.split("/")[-1]}", line {code.co_firstlineno} does not support argument "{var}", so it is ignored. To remove this warning, add **kwargs to the function definition.'
                )

    @staticmethod
    def generator(
        OM: functools.partial, AM: functools.partial
    ) -> Callable[..., complex]:
        """Static method for generating the function creating the scattering matrix element from the compact models

        Args:
            OM (functools.partial): Compact model for Optical Length
            AM (functools.partial): Compact model for Loss

        Returns:
            function: function to crete the element of the scattering matrix.

        """

        def TOT(**kwargs) -> complex:
            OML = (
                Model_from_NazcaCM.call_partial(OM, **kwargs)
                if callable(OM)
                else copy(OM)
            )
            AML = (
                Model_from_NazcaCM.call_partial(AM, **kwargs)
                if callable(AM)
                else copy(AM)
            )
            return 10.0 ** (0.05 * AML) * np.exp(2.0j * np.pi / kwargs.get("wl") * OML)

        return TOT

    @staticmethod
    def wraps(func: functools.partial) -> Callable[..., complex]:
        """Static method for generating the function creating the scattering matrix element from the amplitude
        compact model.

        Args:
            func (functools.partial): Compact model for Amplitude

        Returns:
            function: function to create the element of the scattering matrix.
        """

        def wrapper(**kwargs):
            Inner = (
                Model_from_NazcaCM.call_partial(func, **kwargs)
                if callable(func)
                else copy(func)
            )
            return Inner

        return wrapper

    @classmethod
    def check_init(
        cls,
        cell: nd.Cell,
        ampl_model: str = None,
        loss_model: str = None,
        optlength_model: str = None,
        allowed: dict[str, dict[str, Any]] = None,
    ) -> Model_from_NazcaCM:
        """Alternative Creator. Mainly used for debugging

        This creator will directly try to solve the obtained model before returning.

        Args:
            cls (TYPE): DESCRIPTION.
            cell (TYPE): DESCRIPTION.
            ampl_model (TYPE, optional): DESCRIPTION. Defaults to None.
            loss_model (TYPE, optional): DESCRIPTION. Defaults to None.
            optlength_model (TYPE, optional): DESCRIPTION. Defaults to None.
            allowed (TYPE, optional): DESCRIPTION. Defaults to None.

        Returns:
            Model: Model of the cell.

        Raises:
            RuntimeError: if calling solve on the Model fails.
        """
        try:
            obj = cls(
                cell=cell,
                ampl_model=ampl_model,
                loss_model=loss_model,
                optlength_model=optlength_model,
                allowed=allowed,
            )
            obj.solve(wl=1.55)
            return obj
        except AttributeError:
            raise RuntimeError(f"Model for cell {obj.name} is not solved")

    @classmethod
    def nazca_init(
        cls,
        cell: nd.Cell,
        ampl_model: str = None,
        loss_model: str = None,
        optlength_model: str = None,
        allowed: dict[str, dict[str, Any]] = None,
    ) -> Union[Model_from_NazcaCM, Solver]:
        """Alternative Creator to be used inside Nazca Integration

        This alternative creator will check if a model can be obtained from the Nazca cell. If not, an empty Solver will
        be returned instead. For a more detailed description on how to use this class see the __init__ method.

        Args:
            cell (Nazca Cell): it expects a Nazca cell with some compact model defined.
            ampl_model (str): model to the used to directly get the scattering matrix amplitudes
            loss_model (str): model to be used to estimate the loss
            optlength_model (str): model to be used for the optical length (phase)
            allowed (dict): mapping {Mode:extra}. The allowed mode in the cell and the extra information to
                pass to the compact model to build the optical length.

        Returns:
            Model or Solver: Model of the cell or empy Solver.

        """
        obj = cls(
            cell=cell,
            ampl_model=ampl_model,
            loss_model=loss_model,
            optlength_model=optlength_model,
            allowed=allowed,
        )
        if obj.is_empty():
            logger.debug(f"Model of cell {obj.name} is empy")
            return Solver(name=obj.name)
        else:
            return obj

    def create_S(self) -> np.ndarray:
        """Creates the scattering matrix

        Returns:
            ndarray: Scattering Matrix
        """
        self.S = np.zeros((self.N, self.N), dtype="complex")
        for (pin1, pin2), CM in self.CM.items():
            CM = self.CM[(pin1, pin2)]
            self.S[self.pin_dic[pin1], self.pin_dic[pin2]] = (
                CM(**self.param_dic) if callable(CM) else CM
            )
        return self.S

    def __str__(self):
        """Formatter function for printing"""
        return f"Model (id={id(self)}) from Nazca {self.name:20}"


def get_solver_from_nazca(
    Cell,
    fullreturn=False,
    drc=True,
    prune=True,
    infolevel=0,
    atol=1e-4,
    ampl_model="ampl",
    optlength_model="optlen",
    loss_model="optloss",
    allowed=None,
):
    """Build a Solver Object from a Nazca Cell

    Args:
        Cell (Cell): Nazca cell
        fullreturn (bool): If True, the returned object is a dictionary containing all the solver of all sub-cells in the hierarchy. If False (default) only the top solver is returned.
        drc (bool): if True (default) pin connections violating drc will not be connected
        prune (bool): Remove empty branches from the solver structure. Default is True
        infolevel (int): Regulates the amount of info written to stdout. Default is 0 (no info added). The higher the number the more the information
        atol (flost): absolute tolerance of pin connection
        ampl_model (string): Name of the tracker to get the scattering amplidtude commpact model. It takes precedence over the loss and optical length model.
        optlength_model (string): Name of the tracker to get the optical length compact model. It is only used if ampl_model is not found in the cell.
        loss_model (string): Name of the tracker to get the loss compact model. It is only used if ampl_model is not found in the cell.

    Returns:
        Solver or dict: see fullreturn
    """

    allowed = {"": dict(pol=0, mode=0)} if allowed is None else allowed

    models = {}
    structures = {}
    cells_with_model = []

    it1, it2 = tee(nd.cell_iter(Cell, hierarchy="full", topdown=False))

    for params in it1:
        if params.cell_create:
            if params.cell.auxiliary:
                continue
            if "bbox" in params.cell.cell_name:
                continue
            if "icon" in params.cell.cell_name:
                continue
            if "stub" in params.cell.cell_name:
                continue
            # if set([t.cell.cnode for t in params.branch]).intersection(cells_with_model)!=set(): continue

            if infolevel > 0:
                logger.debug(f"{params.cell.cnode}")
            # logger.debug(params.cell.model_info)
            if params.cell.model_info["model"] is None:
                models[params.cell.cnode] = Model_from_NazcaCM.nazca_init(
                    params.cell,
                    ampl_model=ampl_model,
                    loss_model=loss_model,
                    optlength_model=optlength_model,
                    allowed=allowed,
                )
            else:
                models[params.cell.cnode] = params.cell.model_info["model"]
            if not models[params.cell.cnode].is_empty():
                cells_with_model.append(params.cell.cnode)
                if infolevel > 1:
                    logger.debug(f"  {models[params.cell.cnode]}")
                if infolevel > 0:
                    logger.debug("")
                continue

            params.iters["instance"], itcopy = tee(params.iters["instance"])
            for inode, xya, flip in itcopy:
                if inode.instance.cell.auxiliary:
                    continue
                if "bbox" in inode.instance.cell.cell_name:
                    continue
                if "stub" in inode.instance.cnode.up.cell.cell_name:
                    continue
                if "icon" in inode.instance.cnode.up.cell.cell_name:
                    continue
                param_mapping = (
                    inode.instance.model_info["param_mapping"]
                    if "param_mapping" in inode.instance.model_info
                    else {}
                )
                if infolevel > 1:
                    logger.debug(f"  {inode}")
                if isinstance(models[inode.instance.cell.cnode], Model):
                    ST = Structure(
                        model=models[inode.instance.cell.cnode],
                        param_mapping=param_mapping,
                    )
                if isinstance(models[inode.instance.cell.cnode], Solver):
                    ST = Structure(
                        solver=models[inode.instance.cell.cnode],
                        param_mapping=param_mapping,
                    )
                structures[inode] = ST
                models[params.cell.cnode].add_structure(ST)

            for inode, xya, flip in params.iters["instance"]:
                if inode.instance.cell.auxiliary:
                    continue
                if "bbox" in inode.instance.cell.cell_name:
                    continue
                if "stub" in inode.instance.cnode.up.cell.cell_name:
                    continue
                if "icon" in inode.instance.cnode.up.cell.cell_name:
                    continue
                if infolevel > 0:
                    logger.debug(f"  {inode}")
                for name, pin in inode.instance.pin.items():
                    # if infolevel>1: logger.debug(f'    {name} : {pin}')
                    if name not in structures[inode].get_pin_basenames():
                        continue
                    if name in ["org"]:
                        continue
                    if pin.type == "bbox":
                        continue
                    if "stub" in pin.cnode.up.cell.cell_name:
                        continue
                    if "icon" in pin.cnode.up.cell.cell_name:
                        continue
                    if infolevel > 0:
                        logger.debug(f"    {name} : {pin}")
                    new_nb_geo = []
                    for (tnode, pointer) in pin.nb_geo:
                        if infolevel > 1:
                            logger.debug(f"      {tnode}")
                        if tnode.type == "bbox":
                            continue
                        if tnode.name.endswith("org"):
                            continue
                        if "bbox" in tnode.name:
                            continue
                        if tnode.io is None:
                            continue
                        new_nb_geo.append((tnode, pointer))
                    connected = False
                    for (tnode, pointer) in new_nb_geo:
                        if tnode.name.isdigit() and len(new_nb_geo) > 1:
                            continue
                        if drc == True and pin.xs != tnode.xs:
                            continue
                        xya = list(nd.diff(pin, tnode))
                        xya[2] = xya[2] % 360.0
                        if infolevel > 0:
                            logger.debug(f"      {tnode} : {xya}")
                        mode_out = structures[inode].get_pin_modenames(name)
                        if infolevel > 2:
                            logger.debug(f"      {mode_out}, {xya}")
                        if (
                            np.all(np.isclose(xya, [0.0, 0.0, 180.0], atol=atol))
                            and tnode.up.name
                            in structures[tnode.cnode].get_pin_basenames()
                        ):
                            models[params.cell.cnode].connect_all(
                                structures[inode],
                                name,
                                structures[tnode.cnode],
                                tnode.up.name,
                            )
                            connected = True
                    if connected:
                        continue
                    for (tnode, pointer) in new_nb_geo:
                        if tnode.name.isdigit() and len(new_nb_geo) > 1:
                            continue
                        if drc == True and pin.xs != tnode.xs:
                            continue
                        xya = list(nd.diff(pin, tnode))
                        xya[2] = xya[2] % 360.0
                        if infolevel > 0:
                            logger.debug(f"      {tnode} : {xya}")
                        mode_out = structures[inode].get_pin_modenames(name)
                        if infolevel > 2:
                            logger.debug(f"      {mode_out}, {xya}")
                        if np.all(
                            np.isclose(xya, [0.0, 0.0, 0.0], atol=atol)
                        ) or np.all(np.isclose(xya, [0.0, 0.0, 360.0], atol=atol)):
                            for mi in mode_out:
                                name_out = name if mi == "" else "_".join([name, mi])
                                name_in = (
                                    tnode.name
                                    if mi == ""
                                    else "_".join([tnode.name, mi])
                                )
                                models[params.cell.cnode].map_pins(
                                    {name_in: structures[inode].pin[name_out]}
                                )
                                if infolevel > 1:
                                    logger.debug(f"        {name_in} : {name_out}")
            if infolevel > 0:
                logger.debug("\n")

    if prune:
        for cnode, solver in copy(models).items():
            if solver.prune():
                models.pop(cnode)

    if len(models) == 0:
        nd.main_logger(f"Impossible to buld solver for cell {Cell.name}", "error")
        return None

    if fullreturn:
        return models
    else:
        return models[Cell.cnode]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Mar 31 09:48:24 2021

@author: marco
"""
import functools

import numpy as np
import types


class ProtectedPartial(functools.partial):
    """Like partial, but keywords provided at creation cannot be overwritten al call time"""

    def __call__(self, /, *args, **keywords):
        keywords = {**keywords, **self.keywords}
        return self.func(*self.args, *args, **keywords)


def map_args(func, varsmap={}):
    """This function maps the argument names from a funciton to a new one

    Example: usnign this gunction on f(y, x=2) with mapping {'x' : 'a', 'y' : 'b'} will return f(a, b=2)
    Order of arguments and default values will be preserved

    Args:
        func (function):    Function which arguments need to be mapped
        varsmap (dict):     Dictionary of the mapping {'oldname' : 'newname'}

    Returns:
        function:   New function with the updated argument names
    """
    y = func
    oldargs = y.__code__.co_varnames[: y.__code__.co_argcount]
    new_args = []
    for x in oldargs:
        if x in varsmap:
            new_args.append(varsmap[x])
        else:
            new_args.append(x)
    new_args = tuple(new_args) + y.__code__.co_varnames[y.__code__.co_argcount :]
    print(new_args)
    y_code = types.CodeType(
        y.__code__.co_argcount,
        y.__code__.co_kwonlyargcount,
        y.__code__.co_nlocals,
        y.__code__.co_stacksize,
        y.__code__.co_flags,
        y.__code__.co_code,
        y.__code__.co_consts,
        y.__code__.co_names,  # y.__code__.co_varnames, \
        new_args,
        y.__code__.co_filename,
        y.__code__.co_name,
        y.__code__.co_firstlineno,
        y.__code__.co_lnotab,
    )
    return types.FunctionType(y_code, y.__globals__, y.__name__, y.__defaults__)


def line(N):
    """It returns N-values spaced by 1 and centered around 0"""
    return np.linspace(-0.5 * (N - 1), 0.5 * (N - 1), N)


class GaussianBeam:
    def __init__(
        self,
        w0,
        n,
        wl,
        z0=0.0,
        x0=0.0,
        theta=0.0,
    ):
        """Gaussian beam for propagation in 2D slabs.

        The beam is launched at (z0, x0) at an angle theta with respect to the z-axis and is characterized by
        a beam waist w0. The medium is assumed to be uniform in x and z, so that the power is constant among two
        xy planes.

        Args:
            w0 (float): Beam waist in [um].
            n (float): Refractive index of the medium.
            wl (float): Wavelength in [um].
            z0 (float): Initial z position of the beam.
            x0 (float): Initial x position of the beam.
            theta (float): Launch angle in [deg] with respect to the x-axis.

        """
        self.wl = wl
        self.w0 = w0
        self.n = n
        self.z_r = np.pi * w0 * w0 * n / wl  # Rayleigh range
        self.z0 = z0
        self.x0 = x0
        self.theta = np.deg2rad(-theta)  # Convert to radians here

    def rotate(self, z, x):
        """Creates a 2D rotation matrix based on the angle theta.

        Args:
            z (list | np.array): The z position vector (horizontal).
            x (list | np.array): The x position vector (vertical).

        Returns:
            np.array, np.array: The rotated vectors
        """
        z = z - self.z0
        x = x - self.x0
        _z = np.cos(self.theta) * z - np.sin(self.theta) * x
        _x = np.sin(self.theta) * z + np.cos(self.theta) * x
        return _z, _x

    def waist(self, z):
        """Beam waist.

        Args:
            z (float | list | numpy.array): Position in [um].

        Returns:
            float | list | numpy.array: The beam waist at the position z.
        """
        return self.w0 * np.sqrt(1 + (z / self.z_r) ** 2)

    def curvature(self, z):
        """Curvature radius.

        Args:
            z (float | list | numpy.array): Position in [um].

        Returns:
            float | list | numpy.array: The curvature at the position z.
        """
        return z * (1 + (self.z_r / z) ** 2)

    def gouy(self, z):
        """Gouy phase.

        Args:
            z (float | list | numpy.array): Position in [um].

        Returns:
            float | list | numpy.array: The Gouy phase at the position z.
        """
        return np.arctan(z / self.z_r)

    def field(self, z, x):
        """Complex electric field.

        Args:
            z (float | list | numpy.array): z position in [um].
            x (float): x position in [um].

        Returns:
            np.ndarray: x, z, field
        """
        _z, _x = self.rotate(z, x)

        k_z = 2 * np.pi * self.n / self.wl
        k_x = 2 * np.pi * self.n / self.wl
        _waist = self.waist(_z)

        _amplitude = np.sqrt(np.sqrt(2 / np.pi) / _waist) * np.exp(
            -_x * _x / (_waist * _waist)
        )
        _phase = np.exp(
            -1j * (k_z * _z + k_x * _x * _x / (2 * self.curvature(_z)) - self.gouy(_z))
        )

        _field = _amplitude * _phase

        return _field



from __future__ import annotations
from typing import List, Dict, Tuple, Any, Optional

import numpy as np
import pandas
import lekkersim
from lekkersim.scattering import S_matrix
from copy import deepcopy
from copy import copy
import lekkersim.model as mod
from lekkersim.log import logger


class Structure:
    """Class defining a single element of the photonic circuit

    Args:
        pin_list (list) : list of str containing the names of the pins. Default is empty list
        model (Model) : model object from which the structue may be derived. Defauls is none
        solver (Solver) : solver object from which the structue may be derived. Defauls is none
        param_mapping (dict): dictionary of {oldname (str) : newname (str)} containning the mapping of the names of the parameters. Default is empty dict
    """

    def __init__(
        self,
        pin_list: List[str] = [],
        model: lekkersim.Model = None,
        solver: lekkersim.Solver = None,
        param_mapping: Dict[str, str] = None,
    ) -> None:
        """Creator"""
        self.pin_list = []
        self.pin_dic = {}
        for i, pin in enumerate(pin_list):
            self.pin_list.append((self, pin))
            self.pin_dic[(self, pin)] = i
        self.in_pins = {}
        self.out_pins = {}
        self.structures = []
        self.conn_dict = {}
        self.connected_to = []
        self.gone_to = self
        self.N = len(pin_list)
        self.model = model
        self.sn = 1
        self.param_dic = {}
        if model is not None:
            for pin, i in model.pin_dic.items():
                self.pin_list.append((self, pin))
                self.pin_dic[(self, pin)] = i
        if solver is not None:
            for pin in solver.pin_mapping:
                self.pin_list.append((self, pin))
                self.pin_dic[(self, pin)] = None
        self.solver = solver
        self.param_mapping = {}
        param_mapping = {} if param_mapping is None else param_mapping
        for oldname, newname in param_mapping.items():
            self.param_mapping[newname] = oldname

    def __str__(self) -> str:
        """Formatter for printing"""
        if self.model is not None:
            return f"Structure (id={id(self)}) containing {str(self.model)}"
        elif self.solver is not None:
            return f"Structure (id={id(self)}) containing {str(self.solver)}"
        else:
            return f"{self.__class__.__name__} instance at {id(self)}>"

    @property
    def pin(self) -> Dict[str, Tuple[Structure, str]]:
        """Alllows the feeding of the sintax structure.pin['pin_name'] where the tuple of structure and pin is required

        Returns:
            dict: Dictionary of the pins, in the format {'pinname' : (self, pinname)}
        """
        dic = {pin: (self, pin) for sl, pin in self.pin_list}
        return dic

    @property
    def pin_count(self) -> int:
        """Number of pins in a structure"""
        return len(self.pin_list)

    def get_pin_basenames(self) -> str:
        """Return the set of the basename of the pins"""
        return set([pin[1].split("_")[0] for pin in self.pin_list])

    def get_pin_modenames(self, target) -> List[str]:
        """Return list of mode names given a pin basename"""
        li = []
        for pin in self.pin_list:
            try:
                pinname, modename = pin[1].split("_")
            except ValueError:
                pinname, modename = pin[1], ""
            if pinname == target:
                li.append(modename)
        return li

    def get_pins(self, basename=None) -> List[Tuple[Structure, str]]:
        """Return list of the pins tuple

        Args:
            basename (str): if pins are in the form 'basename_modename', only the ones with the corresponding basename are returned

        Returs:
            list: list of pin tuples
        """
        if basename is None:
            return self.pin_list
        else:
            return [pin for pin in self.pin_list if pin.split("_")[0] == basename]

    def update_params(self, param_dic: Dict[str, Any]) -> None:
        """Updated the parametes dictionary of the represented optic componet

        Args:
            param_dic (dictionary): dcitionary {'param_name':param_value} containing value of the parameters to be updated

        Returns:
            None
        """
        update_dic = deepcopy(param_dic)
        for newname, oldname in self.param_mapping.items():
            if oldname in update_dic:
                update_dic.pop(oldname)
            if newname in param_dic:
                update_dic[oldname] = update_dic.pop(newname)
        # print(self,param_dic,self.param_mapping,update_dic)
        self.param_dic = update_dic
        if self.model is not None:
            self.model.update_params(update_dic)
        if self.solver is not None:
            self.solver.update_params(update_dic)

    def createS(self) -> np.ndarray:
        """Creates the scattering matrix of the components

        Returns:
            ndarray: Scattering Matrix
        """
        if self.model is not None:
            self.Smatrix = self.model.solve(**self.param_dic).create_S()
        if self.solver is not None:
            model = self.solver.solve(**self.param_dic)
            for pin, i in model.pin_dic.items():
                self.pin_dic[(self, pin)] = i
            self.Smatrix = model.create_S()
        self.N = np.shape(self.Smatrix)[-1]
        self.ns = np.shape(self.Smatrix)[0]
        return self.Smatrix

    def print_pindic(self) -> None:
        """Print the mappping between the pins and the entries of the scatterng matrix"""
        for (st, pin), i in self.pin_dic.items():
            print("(%50s, %5s) : %3i" % (st, pin, i))

    def print_conn(self) -> None:
        """Print the connection of this structure to other ones"""
        for (st1, pin1), (st2, pin2) in self.conn_dict.items():
            print("(%50s, %5s) --> (%50s, %5s)" % (st1, pin1, st2, pin2))

    def reset(self) -> None:
        """Reset the mapping that keeps track of the solving"""
        self.gone_to = self
        self.param_dic = {}

    def add_pin(self, pin: Tuple[Structure, str]) -> None:
        """Add pin to structure

        Args:
            pin (tuple) : pin to be added

        Returns:
            None
        """
        if pin in self.pin_list:
            raise Exception("Pin already present, nothing is done")
        else:
            self.pin_list.append(pin)
            self.pin_dic[pin] = self.N
            self.N += 1

    def sel_input(self, pin_list: List[Tuple[Structure, str]]) -> None:
        """Divide pins to be connected providing inputs pins

        Args:
            pin_list (list) : list of tuples (structure (Structure), pin_name (str)) providing input pins

        Returns:
            None
        """
        self.in_list = []
        self.out_list = copy(self.pin_list)
        for pin in pin_list:
            self.in_list.append(pin)
            self.out_list.remove(pin)

    def sel_output(self, pin_list: List[Tuple[Structure, str]]) -> None:
        """Divide pins to be connected providing output pins

        Args:
            pin_list (list) : list of tuples (structure (Structure), pin_name (str)) providing output pins

        Returns:
            None
        """
        self.out_list = []
        self.in_list = copy(self.pin_list)
        # print(self.in_list)
        for pin in pin_list:
            self.out_list.append(pin)
            self.in_list.remove(pin)

    def split_in_out(
        self,
        in_pins: List[Tuple[Structure, str]] = None,
        out_pins: List[Tuple[Structure, str]] = None,
    ) -> None:
        """Created the scattering matrix object with left pins separated from right pins

        Args:
            in_pins (list) : list of tuples (structure (Structure), pin_name (str)) providing "left" pins
            out_pins (list) : list of tuples (structure (Structure), pin_name (str)) providing "right" pins

        Returns:
            None
        """
        # print('Splitting structure: ',self)
        in_pins = self.in_list if in_pins is None else in_pins
        out_pins = self.out_lists if out_pins is None else out_pins

        N = len(in_pins)
        M = len(out_pins)
        self.Sproc = S_matrix(N, M, ns=self.ns)
        for i, p in enumerate(in_pins):
            self.in_pins[p] = i
            for j, q in enumerate(in_pins):
                self.Sproc.S21[:, i, j] = self.Smatrix[
                    :, self.pin_dic[p], self.pin_dic[q]
                ]
        for i, p in enumerate(in_pins):
            for j, q in enumerate(out_pins):
                self.Sproc.S22[:, i, j] = self.Smatrix[
                    :, self.pin_dic[p], self.pin_dic[q]
                ]
        for i, p in enumerate(out_pins):
            self.out_pins[p] = i
            for j, q in enumerate(in_pins):
                self.Sproc.S11[:, i, j] = self.Smatrix[
                    :, self.pin_dic[p], self.pin_dic[q]
                ]
        for i, p in enumerate(out_pins):
            for j, q in enumerate(out_pins):
                self.Sproc.S12[:, i, j] = self.Smatrix[
                    :, self.pin_dic[p], self.pin_dic[q]
                ]

    def get_S_back(self) -> None:
        """Recreates the scattering matrix as ndarray"""
        self.Smatrix = np.concatenate(
            [
                np.concatenate([self.Sproc.S21, self.Sproc.S22], axis=-1),
                np.concatenate([self.Sproc.S11, self.Sproc.S12], axis=-1),
            ],
            axis=-2,
        )
        N = self.Sproc.N
        for pin, i in self.in_pins.items():
            self.pin_dic[pin] = i
        for pin, i in self.out_pins.items():
            self.pin_dic[pin] = i + N
        del self.Sproc
        self.in_pins = {}
        self.out_pins = {}

    def print_pins(self) -> None:
        """Print all pins of the structures, divided in self pins and contained pins"""
        print("Self pins:")
        for c, pinname in self.pin_list:
            if c is self:
                print(pinname)
        if len(self.structures) == 0:
            print("No additional Pins")
        else:
            print("Additional Pins:")
            for cc in self.structures:
                for c, pinname in self.pin_list:
                    if c is cc:
                        print(c, pinname)

    def add_conn(
        self, pin: str, target: Structure, target_pin: Tuple[Structure, str]
    ) -> None:
        """Add connection between a self pin and a pin in another structure

        Args:
            pin (str) : name of self pin
            target (Structure) : structure to which to connect
            target_pin (str) : name of the pin in the target structure to whcih to connect

        Returns:
            None
        """
        tup = self.conn_dict.get((self, pin))
        if tup is not None:
            if tup != (target, target_pin):
                raise Exception("Pin already connected")
        else:
            self.conn_dict[(self, pin)] = (target, target_pin)
        if target not in self.connected_to:
            self.connected_to.append(target)

    def remove_pin(self, pin: str) -> None:
        """Remove pin from structure

        Args:
            pin (str) : name of pin to remove

        Returns:
            None
        """
        if (self, pin) in self.conn_dict:
            self.conn_dict.pop((self, pin))
        else:
            raise Exception(f"Pin {pin} not in conn_dict")
        if (self, pin) in self.pin_list:
            self.pin_list.remove((self, pin))
            self.pin_dic.pop((self, pin))
        else:
            raise Exception(f"Pin {pin} not in conn_dict")

    def cut_connections(self, target: Structure) -> None:
        """Remove all connection to target structure. Pins in self are kept

        Args:
            target (Structure) : structore to which to remove connnetions

        Returns:
            None
        """
        if target not in self.connected_to:
            raise Exception(
                f"Structure {target} is not connected to {self}. Impossible to remove."
            )
        self.connected_to.remove(target)
        copy_dic = copy(self.conn_dict)
        for (s, pin), (t, tpin) in copy_dic.items():
            if t is target:
                if (s, pin) in self.conn_dict:
                    self.conn_dict.pop((s, pin))
                else:
                    raise Exception(f"Pin {pin} not in conn_dict")

    def remove_connections(self, target: Structure) -> None:
        """Remove all connection to target structure. Remove also from self all pins involved in the connections

        Args:
            target (Structure) : structore to which to remove connnetions

        Returns:
            None
        """
        if target not in self.connected_to:
            raise Exception(
                f"Structure {target} is not connected to {self}. Impossible to remove."
            )
        self.connected_to.remove(target)
        copy_dic = copy(self.conn_dict)
        for (s, pin), (t, tpin) in copy_dic.items():
            if t is target:
                self.remove_pin(pin)

    def get_out_to(self, st: Structure) -> List[Tuple[Structure, str]]:
        """Find pins of self with are connected to a target structure

        Args:
            st (Structure) : target structure

        Returns:
            pin_list : list of tuple (structure (Structure), pin_name (str)) contaning the list of pins which connects to the target structure
        """
        pin_list = []
        target_list = [st] + st.structures
        for (loc_c, loc_name), (tar_c, tar_name) in self.conn_dict.items():
            if tar_c in target_list:
                pin_list.append((loc_c, loc_name))
        return pin_list

    def get_in_from(self, st: Structure) -> List[Tuple[Structure, str]]:
        """Find pins of self with are connected from a target structure

        Args:
            st (Structure) : target structure

        Returns:
            pin_list : list of tuple (structure (Structure), pin_name (str)) contaning the list of pins which connects from the target structure
        """
        pin_list = []
        loc_list = [self] + self.structures
        for (source_c, source_name), (loc_c, loc_name) in st.conn_dict.items():
            if loc_c in loc_list:
                pin_list.append((loc_c, loc_name))
        return pin_list

    def join(self, st: Structure) -> Structure:
        """Join two structures to create the one cotaining the merged structure

        Args:
            st (Structure) : target structure to join to self

        Returns:
            Structure : new structure combining the two starting one
        """
        self.createS()
        st.createS()
        # safety checks
        if st in self.structures:
            raise Exception("Cannot add structure: already containded")
        for st1 in st.structures:
            if st1 in self.structures:
                raise Exception("Cannot add structure: already containded")
        # retriving list of pins
        new_st = Structure()
        loc_out = self.get_out_to(st)
        tar_in = st.get_in_from(self)
        # print(loc_out)
        # print(tar_in)
        # Checking connectivity
        if len(loc_out) != len(tar_in):
            raise Exception("Connectivity problem: Different number of pins")
        for pin1 in loc_out:
            if pin1 != st.conn_dict[self.conn_dict[pin1]]:
                raise Exception("Connectivity problem: Not Symmetric")
        # Correct ordering of connection pins
        tar_in = []
        for pin in loc_out:
            tar_in.append(self.conn_dict[pin])

        # getting list of pins for local and target
        # print(loc_out)
        self.sel_output(loc_out)
        st.sel_input(tar_in)
        # print('Structure:',self)
        # print('In pins:',self.in_list)
        # print('Out pins:',self.out_list)

        # print('Structure:',st)
        # print('In pins:',st.in_list)
        # print('Out pins:',st.out_list)

        # Splittig scattering matrces of local and target
        self.split_in_out(self.in_list, self.out_list)
        st.split_in_out(st.in_list, st.out_list)

        # print(self.Sproc)
        # self.Sproc.S_print()

        # Joining S matrices for new one
        new_st.Sproc = self.Sproc.add(st.Sproc)
        new_st.in_pins = self.in_pins
        new_st.out_pins = st.out_pins
        new_st.get_S_back()

        # Generating new structures list
        if len(self.structures) == 0:
            new_st.structures.append(self)
            self.gone_to = new_st
        else:
            for st1 in self.structures:
                new_st.structures.append(st1)
                st1.gone_to = new_st
        if len(st.structures) == 0:
            new_st.structures.append(st)
            st.gone_to = new_st
        else:
            for st1 in st.structures:
                new_st.structures.append(st1)
                st1.gone_to = new_st
        # generating new pin lists
        add_pins = self.pin_list + st.pin_list
        for pin in loc_out + tar_in:
            add_pins.remove(pin)
        for pin in add_pins:
            new_st.add_pin(pin)

        # generating new connected to
        for st1 in self.connected_to + st.connected_to:
            if (st1 not in new_st.connected_to) and (st1 not in new_st.structures):
                new_st.connected_to.append(st1)

        # upgrading connection_dic
        for (st_source, pin_source), (st_target, pin_target) in {
            **self.conn_dict,
            **st.conn_dict,
        }.items():
            if not (
                (st_source in new_st.structures) and (st_target in new_st.structures)
            ):
                new_st.conn_dict[(st_source, pin_source)] = (st_target, pin_target)

        if self.ns == st.ns:
            new_st.ns = self.ns

        return new_st

    def intermediate(
        self, st: Structure, pin_mapping: Dict[str, Tuple[Structure, str]]
    ) -> None:
        """Used to generate the function for monitoring the modes between two structures

        This function is use to generate tge function to coumpe the modes amplitude at selected internal ports between 2 structures.

        Args:
            st (Structure) : structure towards which to monitor the ports. Connectivity between the structures has to be already specified.
            pin_mapping (dict): dictionary of pin mapping in the form {pin_name (str) : (structure (Structure), pin (str) )}.
                used when building the function to keep track of the correct pin naming

        Returns:
            Function: function for returnin modal coefficient given a dictionary with the inputs
            dict: dictionary containing the mapping between ports of st and scattering matrix entry
        """

        loc_out = self.get_out_to(st)
        tar_in = st.get_in_from(self)
        if len(loc_out) != len(tar_in):
            raise Exception("Connectivity problem: Different number of pins")
        for pin1 in loc_out:
            if pin1 != st.conn_dict[self.conn_dict[pin1]]:
                raise Exception("Connectivity problem: Not Symmetric")
        # Correct ordering of connection pins
        tar_in = []
        for pin in loc_out:
            tar_in.append(self.conn_dict[pin])

        self.sel_output(loc_out)
        st.sel_input(tar_in)

        # Splittig scattering matrces of local and target
        self.split_in_out(self.in_list, self.out_list)
        st.split_in_out(st.in_list, st.out_list)

        def solve_inter(dic: Dict[str, complex]) -> Tuple[np.ndarray]:
            """This function calculates the couefficient of the internal modes

            Args:
                dic (dict): dictionary {pin_name (str) : input_amplitude (complex)}. Dictionary containing the complex amplitudes at each input port. Missing port are assumed wiht amplitude 0.0

            Returs:
                numpy array: array of the coefficient of the input modes
                numpy array: array of the coefficient of the output modes
            """
            u = np.zeros(len(self.in_list), dtype=complex)
            d = np.zeros(len(st.out_list), dtype=complex)
            for name, value in dic.items():
                mapped = pin_mapping[name]
                if mapped in self.in_list:
                    u[self.in_pins[mapped]] = value
                if mapped in st.out_list:
                    d[st.out_pins[mapped]] = value
            uo, do = self.Sproc.int_complete(st.Sproc, u, d)
            return uo, do

        return solve_inter, st.in_pins

    def get_model(
        self,
        pin_mapping: Optional[Dict[str, Tuple[Structure, str]]] = None,
        name: Optional[str] = None,
    ):
        """Retunrn model corresponding to structure

        Args:
            pin_mapping (dict) : dictionary contain the information of pin mapping {pin_name (str) : (structure (Structure), pin (str))}
            name (str): name to assign at the solved model

        Returns:
            SolvedModel : model object containing the scattering matrix od the structure
        """
        Smod = np.zeros((self.ns, self.N, self.N), complex)
        if pin_mapping is None:
            pin_mapping = self.solver.pin_mapping
        # if len(pin_mapping)!=len(self.pin_dic):
        # raise Exception('Not all pins mapped correctly')
        #    logger.warning(f'{self}:Running solve without complete mapping: some pins will not be not accessible')
        # for i,pin_name in enumerate(pin_mapping):
        #    pin_dic[pin_name]=i
        #   #print(i,pin_name,pin_mapping[pin_name])
        #    for j,pin_namej in enumerate(pin_mapping):
        #        Smod[:,i,j]=self.Smatrix[:,self.pin_dic[pin_mapping[pin_name]],self.pin_dic[pin_mapping[pin_namej]]]
        pin_dic = {name: self.pin_dic[pin] for name, pin in pin_mapping.items()}
        MOD = mod.SolvedModel(
            pin_dic=pin_dic,
            param_dic=self.param_dic,
            Smatrix=self.Smatrix,
            name=name,
        )
        return MOD

    def raise_pins(self, pini: List[str] = None, pino: List[str] = None):
        """Raises  some pins of the structure into the solver

        Args:
            pini (list): List of the pins to be raised. If None, all the the pins are raised.
            pino (list): List of the names of raised pins. If none, the names are preserved.

        Returns:
            None

        Raises:
            RuntimeError: if trying to raise pins on a composite structure
            ValueError: if len(pini)!=len(pino)
        """
        if pini is None:
            pini = []
            for st, pin in self.pin_list:
                if st != self:
                    raise RuntimeError("Cannnot raise pins of composite structure")
                pini.append(pin)
        pino = pini if pino is None else pino
        if len(pini) != len(pino):
            raise ValueError("pini and pino have different lengths")
        pin_mapping = {_pino: (self, _pini) for _pino, _pini in zip(pino, pini)}
        lekkersim.sol_list[-1].map_pins(pin_mapping)

    # def return_model(self):
    #    return self.model

    # def get_T(self,pin1,pin2):
    #    self.createS()
    #    return np.abs(self.Smatrix[self.pin_dic[(self,pin1)],self.pin_dic[(self,pin2)]])**2.0

    # def get_output(self,input_dic,power=True):
    #    try:
    #        ret=self.model.get_output(input_dic,power=power)
    #        return ret
    #    except AttributeError:
    #        raise ValueError('This structure does not have a model')



# -------------------------------------------
#
# GenSol Package
#
# Python tool for simulation of abstract photonic circuits
#
# @author: Marco Passoni
#
# -------------------------------------------
from __future__ import annotations
import numpy as np
import numpy.linalg as linalg

import lekkersim


class S_matrix:
    """Class implmenting the scattering matrix object and recursion alghoritms for joining two of them"""

    def __init__(self, N: int, M: int, ns: int = None) -> None:
        """Creator

        Args:
            N (int) : number of "left" ports
            M (int) : number of "right" ports
        """
        self.N = N
        self.M = M
        if ns is None:
            self.S11 = np.zeros((M, N), complex)
            self.S22 = np.zeros((N, M), complex)
            self.S12 = np.zeros((M, M), complex)
            self.S21 = np.zeros((N, N), complex)
        else:
            self.ns = ns
            self.S11 = np.zeros((ns, M, N), complex)
            self.S22 = np.zeros((ns, N, M), complex)
            self.S12 = np.zeros((ns, M, M), complex)
            self.S21 = np.zeros((ns, N, N), complex)

    # OLD RECURSION VERSION
    # def add(self,s):
    #    T1=np.matmul(linalg.inv(np.identity(self.N,complex)-np.matmul(self.S12,s.S21)),self.S11)
    #    T2=np.matmul(linalg.inv(np.identity(self.N,complex)-np.matmul(s.S21,self.S12)),s.S22)
    #    self.S11=np.matmul(s.S11,T1)
    #    self.S12=s.S12+np.matmul(np.matmul(s.S11,self.S12),T2)
    #    self.S21=self.S21+np.matmul(np.matmul(self.S22,s.S21),T1)
    #    self.S22=np.matmul(self.S22,T2)

    # NEW RECURSION VERSION
    def add(self, s: S_matrix) -> S_matrix:
        """Recursion algorith for joining two matrices

        Args:
            s (S_matrix) : target S_matrix to join to self

        Returns:
            S_matrix : joined scattering matrix
        """
        if self.M != s.N:
            raise Exception(
                "Trying to concatenate matrices with different intermediate dimension"
            )
        I = np.identity(self.M, complex)
        T1 = np.matmul(s.S11, linalg.inv(I - np.matmul(self.S12, s.S21)))
        T2 = np.matmul(self.S22, linalg.inv(I - np.matmul(s.S21, self.S12)))
        S = S_matrix(self.N, s.M)
        S.S21 = self.S21 + np.matmul(np.matmul(T2, s.S21), self.S11)
        S.S11 = np.matmul(T1, self.S11)
        S.S12 = s.S12 + np.matmul(np.matmul(T1, self.S12), s.S22)
        S.S22 = np.matmul(T2, s.S22)
        return S

    def S_print(self, i: int = None, j: int = None) -> None:
        """Print scattering matrix as numpy array

        Args:
            i,j (int) : number of ports to print. Default is None (all matrix is printed)
        """
        if i == None:
            S = np.vstack(
                [np.hstack([self.S11, self.S12]), np.hstack([self.S21, self.S22])]
            )
        else:
            j = i if j == None else j
            S = np.vstack(
                [
                    np.hstack([self.S11[i, j], self.S12[i, j]]),
                    np.hstack([self.S21[i, j], self.S22[i, j]]),
                ]
            )
        print(S)

    def det(self) -> float:
        """Calculated determinat of scattering matrix"""
        return linalg.det(
            np.vstack(
                [np.hstack([self.S11, self.S12]), np.hstack([self.S21, self.S22])]
            )
        )

    def matrix(self) -> np.ndarray:
        """Return scattering matrix as ndarray

        Returns:
            ndarray : scattering matrix
        """
        return np.vstack(
            [np.hstack([self.S11, self.S12]), np.hstack([self.S21, self.S22])]
        )

    def int_complete(self, S2: S_matrix, u: np.ndarray, d: np.ndarray):
        """Function for computing the internal modal coefficients between two structures

        Args:
            S2 (S_matrix): Second Scattering Matrix
            u (numpy array): coefficients of input of first scattering matrix
            d (numpy array): coefficients of input of second scattering matrix

        Returns:
            numpy array: coefficients of intermediate modes (first to second)
            numpy array: coefficients of intermediate modes (second to first)
        """
        ID = np.identity(self.M)
        ut = (
            np.matmul(self.S11, np.expand_dims(u, -1))
            if len(u) > 0
            else np.zeros((self.M, 1))
        )
        dt = (
            np.matmul(S2.S22, np.expand_dims(d, -1))
            if len(d) > 0
            else np.zeros((self.M, 1))
        )

        uo = linalg.solve(
            ID - np.matmul(self.S12, S2.S21), ut + np.matmul(self.S12, dt)
        )
        do = linalg.solve(ID - np.matmul(S2.S21, self.S12), dt + np.matmul(S2.S21, ut))

        return (np.squeeze(uo, -1), np.squeeze(do, -1))



# -------------------------------------------
#
# GenSol Package
#
# Python tool for simulation of abstract photonic circuits
#
# @author: Marco Passoni
#
# -------------------------------------------


"""File containing the model calls and related methods
"""
from __future__ import annotations
import functools
from typing import Any, List, Dict, Tuple, Callable, TYPE_CHECKING, Union, Optional, Set

from copy import deepcopy
from copy import copy
import yaml
import io
import datetime

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.axes
import pandas as pd
from scipy.interpolate import interp1d, LinearNDInterpolator
from scipy.integrate import quad_vec

import lekkersim.structure
from lekkersim import sol_list
from lekkersim import logger
import lekkersim
import lekkersim.log
from lekkersim.utils import line, GaussianBeam, ProtectedPartial


def diag_blocks(array_list: List[np.ndarray]) -> np.ndarray:
    """Function building a block diagonal array for list of array

    Args:
        array_list (list): list of ndarrays. Each array is a 2D square array.

    Returns:
         ndarray: 2D array with the arrays in array_list as blocks in the diagonal
    """
    for A in array_list:
        if np.shape(A)[0] != np.shape(A)[1]:
            raise ValueError("Matrix is not square")
    Nl = [np.shape(A)[0] for A in array_list]
    N = sum(Nl)
    M = np.zeros((N, N), dtype=complex)
    m = 0
    for n, A in zip(Nl, array_list):
        M[m : m + n, m : m + n] = A
        m += n
    return M


class Model:
    """Class Model

    It contains the model (definition of the scattering matrix) for each base photonic building block
    This is the general template for the class, no model is defined here.
    Each model is a separate class based on the main one.
    The only use case for using this class directly is to call it without arguments to create an empy
    model that will be eliminated if prune is called

    """

    def __init__(
        self,
        pin_dic: Dict[str, int] = None,
        param_dic: Dict[str, Any] = None,
        Smatrix: np.ndarray = None,
    ) -> None:
        """Initialize the model

        Args:
            pin_dic (dictionary): list of strings containing the model's pin names
            param_dic (dictionary): dictionary {'param_name':param_value} containing the definition of
                the model's parameters.
            Smatrix (ndarray) : Fixed S_matrix of the model

        """
        self.pin_dic = {} if pin_dic is None else pin_dic
        self.N = len(self.pin_dic)
        if Smatrix is not None:
            if self.N != np.shape(Smatrix)[-1]:
                self.N = np.shape(Smatrix)[-1]
        self.S = np.identity(self.N, complex) if Smatrix is None else Smatrix
        self.param_dic = {} if param_dic is None else param_dic
        self.default_params = deepcopy(self.param_dic)

    def is_empty(self) -> bool:
        """Checks if model is empy

        Returns:
            Bool: False is model has no pins, True otherwise

        """
        if len(self.pin_dic) > 0:
            return False
        else:
            return True

    def _expand_S(self) -> np.ndarray:
        self.N = self.N // self.np
        S = self._create_S()
        self.N = self.N * self.np
        # self.S= diag_blocks(self.np*[S])
        return diag_blocks(self.np * [S])

    def inspect(self) -> None:
        """Function that print self. It mocks the same function of Solver"""
        print(f"{self}")

    def create_S(self) -> np.ndarray:
        """Function for returning the scattering matrix of the model

        Returns:
            ndarray: Scattering matrix of the model
        """
        return self.S

    def print_S(self, func: Callable = np.abs) -> None:
        """Function for nice printing of scattering matrix in agreement with pins

        Args:
            func (callable): function to be applied at the scattering matrix before returning.
                if None (default), the raw complex coefficients are printed
        """
        func = (lambda x: x) if func is None else func
        a = list(self.pin_dic.keys())
        ind = list(self.pin_dic.values())
        indsort = np.argsort(a)
        a = [a[i] for i in indsort]
        indsort = np.array(ind)[indsort]
        for pin, i in self.pin_dic.items():
            print(pin, i)
        S = self.create_S()
        I, J = np.meshgrid(indsort, indsort, indexing="ij")
        S = S[0, I, J] if len(np.shape(S)) == 3 else S[I, J]
        S = func(S) if func is not None else S
        st = "            "
        for p in a:
            st += f" {p:8} "
        st += "\n"
        for i, pi in enumerate(a):
            st += f" {pi:8} "
            for j, pj in enumerate(a):
                pr = S[i, j]
                st += f" {pr:8.4f} "
            st += "\n"
        print(st)

    def S2PD(self, func: Callable = None) -> pd.DataFrame:
        """Function for returning the Scattering Matrix as a PD Dataframe

        Args:
            func (callable): function to be applied at the scattering matrix before returning.
                if None (default), the raw complex coefficients are returned

        Returns:
            Pandas DataFrame: Scattering Matrix with name of pins
        """
        a = list(self.pin_dic.keys())
        ind = list(self.pin_dic.values())
        indsort = np.argsort(a)
        a = [a[i] for i in indsort]
        indsort = np.array(ind)[indsort]
        S = self.create_S()
        I, J = np.meshgrid(indsort, indsort, indexing="ij")
        S = S[0, I, J] if len(np.shape(S)) == 3 else S[I, J]
        S = func(S) if func is not None else S
        data = pd.DataFrame(data=S, index=a, columns=a)
        return data

    def get_T(self, pin1: str, pin2: str) -> float:
        """Function for returning the energy transmission between two ports

        If the two ports are the same the returned value has the meaning of reflection

        Args:
            pin1 (str): Name of input pin
            pin2 (str): Name of output pin

        Returns:
            float: Energy transmission between the ports
        """
        if np.shape(self.S)[0] > 1:
            logger.warning(
                f"{self}:Using get_T on a sweep solve. Consider using get_data"
            )
        return np.abs(self.S[0, self.pin_dic[pin1], self.pin_dic[pin2]]) ** 2.0

    def get_PH(self, pin1: str, pin2: str) -> float:
        """Function for returning the phase of the transmission between two ports

        If the two ports are the same the returned value has the meaning of reflection

        Args:
            pin1 (str): Name of input pin
            pin2 (str): Name of output pin

        Returns:
            float: Phase of the transmission between the ports
        """
        if np.shape(self.S)[0] > 1:
            logger.warning(
                f"{self}:Using get_PH on a sweep solve. Consider using get_data"
            )
        return np.angle(self.S[0, self.pin_dic[pin1], self.pin_dic[pin2]])

    def get_A(self, pin1: str, pin2: str) -> complex:
        """Function for returning complex amplitude of the transmission between two ports

        Args:
            pin1 (str): Name of input pin
            pin2 (str): Name of output pin

        Returns:
            float: Complex amplitude of the transmission between the ports
        """
        if np.shape(self.S)[0] > 1:
            logger.warning(
                f"{self}:Using get_A on a sweep solve. Consider using get_data"
            )
        return self.S[0, self.pin_dic[pin1], self.pin_dic[pin2]]

    def expand_mode(self, mode_list: List[str]):
        """This function expands the model by adding additional modes.

        For each pin a number of pins equal the length of mode_list will be created. The pin names will be "{pinname}_{modename}}".
        Each mode will have the same behavior.

        Args:
            mode_list (list) : list of strings containing the modenames.

        Returns:
            Model : new model with expanded modes
        """
        self.np = len(mode_list)
        self.mode_list = mode_list
        new_pin_dic = {}
        for name, n in self.pin_dic.items():
            for i, mode in enumerate(self.mode_list):
                new_pin_dic[f"{name}_{mode}"] = i * self.N + n
        self.pin_dic = new_pin_dic
        self._create_S = self.create_S
        self.create_S = self._expand_S
        self.N = self.N * self.np
        return self

    def get_output(
        self, input_dic: Dict[str, float], power: bool = True
    ) -> Dict[str, float]:
        """Returns the outputs from all ports of the model given the inputs amplitudes

        Args:
            input_dic (dict): dictionary {pin_name (str) : input_amplitude (complex)}.
                Dictionary containing the complex amplitudes at each input port. Missing port are assumed
                with amplitude 0.0.
            power (bool): If True, returned values are power transmissions. If False, complex amplitudes
                are instead returned. Default is True.

        Returns:
            dict: Dictionary containing the outputs in the form {pin_name (str) : output (float or complex)}
        """
        if np.shape(self.S)[0] > 1:
            logger.warning(
                f"{self}:Using get_output on a sweep solve. Consider using get_full_output"
            )
        l1 = list(self.pin_dic.keys())
        l2 = list(input_dic.keys())
        for pin in l2:
            l1.remove(pin)
        if l1 != []:
            for pin in l1:
                input_dic[pin] = 0.0 + 0.0j
        u = np.zeros(self.N, complex)
        for pin, i in self.pin_dic.items():
            u[i] = input_dic[pin]
        d = np.dot(self.S[0, :, :], u)
        out_dic = {}
        for pin, i in self.pin_dic.items():
            if power:
                out_dic[pin] = np.abs(d[i]) ** 2.0
            else:
                out_dic[pin] = d[i]
        return out_dic

    def put(
        self, source_pin: str = None, target_pin=None, param_mapping={}
    ) -> lekkersim.Structure:
        """Function for putting a model in a Solver object, and eventually specify connections

         This function creates a Structure object for the model and place it in the current active Solver
         If both pins and pint are provided, the connection also is made.

        Args:
             source_pin (str): pin of model to be connected
             target_pin (tuple): tuple (structure (Structure) , pin (str)) existing structure and pin to
                which to connect pins of model
             param_mapping (dict): dictionary of {oldname (str) : newname (str)} containing the mapping
                of the names of the parameters

         Returns:
             Structure: the Structure instance created from the model
        """
        # ST=solver.structure.Structure(model=deepcopy(self),param_mapping=param_mapping)
        ST = lekkersim.structure.Structure(model=self, param_mapping=param_mapping)
        sol_list[-1].add_structure(ST)
        if (source_pin is not None) and (target_pin is not None):
            sol_list[-1].connect(ST, source_pin, target_pin[0], target_pin[1])
        return ST

    def solve(self, **kargs) -> lekkersim.SolvedModel:
        """Function for returning the solved model

        This function is to align the behavior of the Model and Solver class.

        Args:
            kwargs: dictionary of the parameters {param_name (str) : param_value (usually float)}

        Returns:
            model: solved model of self
        """
        logger.debug(f"Solving Model {self}")
        self.param_dic.update(self.default_params)
        ns = 1
        for name in kargs:
            kargs[name] = np.reshape(kargs[name], -1)
            if len(kargs[name]) == 1:
                continue
            if ns == 1:
                ns = len(kargs[name])
            else:
                if ns != len(kargs[name]):
                    raise Exception("Different lengths between parameter arrays")
        up_dic = {}
        S_list = []
        for i in range(ns):
            for name, values in kargs.items():
                up_dic[name] = values[0] if len(values) == 1 else values[i]
            self.param_dic.update(up_dic)
            S_list.append(self.create_S())
        return SolvedModel(
            pin_dic=self.pin_dic,
            param_dic=kargs,
            Smatrix=np.array(S_list),
            name=type(self).__name__,
        )

    def show_free_pins(self) -> None:
        """Function for printing pins of model"""
        print(f"Pins of model {self} (id={id(self)})")
        for pin, n in self.pin_dic.items():
            print(f"{pin:5s}:{n:5}")
        print("")

    def pin_mapping(self, pin_mapping: Dict[str, str]):
        """Function for changing the names of the pins of a model

        Args:
            pin_mapping (dict): Dictionary containing the mapping of the pin names.
                Format is {'oldname' : 'newname'}

        Returns:
            model: model of self with updated pin names
        """
        for pin in copy(self.pin_dic):
            if pin in pin_mapping:
                n = self.pin_dic.pop(pin)
                self.pin_dic[pin_mapping[pin]] = n
        return self

    def update_params(self, update_dic: Dict[str, Any]) -> None:
        """Update the parameters of model, setting defaults when value is not provides

        Args:
            update_dic (dict) : dictionary of parameters in the from
                {param_name (str) : param_value (usually float)}
        """
        self.param_dic.update(self.default_params)
        self.param_dic.update(update_dic)

    def prune(self) -> None:
        """Check if the model is empty

        Returns:
            bool: True if the model is empty, False otherwise
        """
        return self.pin_dic == {}

    def get_pin_modes(self, pin: str) -> List[Tuple[str, str]]:
        """Parse the pins for locating the pins with the same base name

        Assumes for the pins a name in the form pinname_modename.

        Args:
            pin (str): base name of the pin

        Returns:
            list: list of modenames for which pinname==pin
        """
        li = []
        for _pin in self.pin_dic:
            try:
                pinname, modename = _pin.split("_")
            except ValueError:
                pinname, modename = _pin, ""
            if pinname == pin:
                li.append((pinname, modename))
        return li

    def get_pin_basenames(self) -> List[str]:
        """Returns a list of the basenames of the pins"""
        pin_basenamens = set([pinname.split("_")[0] for pinname in self.pin_dic])
        return list(pin_basenamens)

    def __str__(self):
        """Formatter function for printing"""
        return f"Model object (id={id(self)}) with pins: {list(self.pin_dic)}"


class SolvedModel(Model):
    """Class for storing data of a solver mode.

    Do not use this class directly. It is returned from all solve methods. It is convenient for extracting data
    """

    def __init__(
        self,
        pin_dic: Optional[Dict[str, int]] = None,
        param_dic: Optional[Dict[str, Any]] = None,
        Smatrix: Optional[np.ndarray] = None,
        int_func: Optional[Callable] = None,
        monitor_mapping: Optional[Dict[str, Tuple[lekkersim.Structure, str]]] = None,
        name: Optional[str] = None,
    ) -> None:
        """Initialize the model

        Args:
            pin_dic (dictionary): list of strings containing the model's pin names
            param_dic (dictionary): dictionary {'param_name':param_value} containing the definition of
                the model's parameters.
            Smatrix (ndarray) : Fixed S_matrix of the model
            int_func (Callable): function for returning the modal coefficient in between two part of the scattering matrix
            monitor_mapping (Dict): Dictionry mapping the name of the monitor to the connected pin
            name (str): name of the solved model

        """
        super().__init__(pin_dic=pin_dic, param_dic=param_dic, Smatrix=Smatrix)
        self.solved_params = deepcopy(param_dic)
        self.ns = np.shape(Smatrix)[0]
        self.int_func = int_func
        self.monitor_mapping = {} if monitor_mapping is None else monitor_mapping
        self.name = name

    def set_intermediate(
        self,
        int_func: Callable,
        monitor_mapping: Dict[str, Tuple[lekkersim.Structure, str]],
    ):
        """Methods for setting the function and mapping for monitors

        Args:
            int_func (callable): Function linking the modal amplitudes at the monitor port to the
                inputs amplitudes
            monitor_mapping (dict): Dictionary connecting the name of the monitor ports with the
                index in the amplitude arrays
        """
        self.int_func = int_func
        self.monitor_mapping = monitor_mapping

    def get_data(self, pin1: str, pin2: str) -> pd.DataFrame:
        """Function for returning transmission data between two ports

        Args:
            pin1 (str): name of the input pin
            pin2 (str): name of the output pin

        Returns:
            pandas DataFrame: Dataframe containing the data. It contains one column per parameter
                given to solve, plus the following:
                    'T'         : Transmission in absolute units
                    'dB'        : Transmission in dB units
                    'Phase'     : Phase of the transmission
                    'Amplitude' : Complex amplitude of the transmission

        """
        params = {}
        if self.ns == 1:
            params = deepcopy(self.solved_params)
        else:
            for name, values in self.solved_params.items():
                if len(values) == 1:
                    params[name] = np.array([values[0] for i in range(self.ns)])
                elif len(values) == self.ns:
                    params[name] = values
                else:
                    raise Exception("Not able to convert to pandas")
        params["T"] = np.abs(self.S[:, self.pin_dic[pin1], self.pin_dic[pin2]]) ** 2.0
        params["dB"] = 20.0 * np.log10(
            np.abs(self.S[:, self.pin_dic[pin1], self.pin_dic[pin2]])
        )
        params["Phase"] = np.angle(self.S[:, self.pin_dic[pin1], self.pin_dic[pin2]])
        params["Amplitude"] = self.S[:, self.pin_dic[pin1], self.pin_dic[pin2]]
        pan = pd.DataFrame.from_dict(params)
        return pan

    def get_full_output(
        self, input_dic: Dict[str, float], power: bool = True
    ) -> pd.DataFrame:
        """Function for getting the output do the system given the inputs

        Args:
            input_dic (dict): Dictionary of the input amplitudes. Format is
                {'pinname' : amplitude (float or complex)}. Missing pins are  assumed to have 0 amplitude.
            power (bool): if True, power (in absolute units) between the ports is returned, otherwise the
                complex amplitude is returned. Default is True

        Returns:
            pandas DataFrame: DataFrame with the outputs. It has one column for each parameter
                given to solve plus one column for each pin.
        """
        params = {}
        if self.ns == 1:
            params = deepcopy(self.solved_params)
        else:
            for name, values in self.solved_params.items():
                if len(values) == 1:
                    params[name] = np.array([values[0] for i in range(self.ns)])
                elif len(values) == self.ns:
                    params[name] = values
                else:
                    raise Exception("Not able to convert to pandas")

        l1 = list(self.pin_dic.keys())
        l2 = list(input_dic.keys())
        for pin in l2:
            l1.remove(pin)
        if l1 != []:
            for pin in l1:
                input_dic[pin] = 0.0 + 0.0j
        u = np.zeros(self.N, complex)
        for pin, i in self.pin_dic.items():
            u[i] = input_dic[pin]

        output = np.matmul(self.S, u)

        for pin, i in self.pin_dic.items():
            params[pin] = np.abs(output[:, i]) ** 2.0 if power else output[:, i]
        pan = pd.DataFrame.from_dict(params)
        return pan

    def get_full_data(self) -> pd.DataFrame:
        """Returns the scattering matrix for all the solved parametes in form of padas DataFrame"""
        params = {}
        if self.ns == 1:
            params = deepcopy(self.solved_params)
        else:
            for name, values in self.solved_params.items():
                if len(values) == 1:
                    params[name] = np.array([values[0] for i in range(self.ns)])
                elif len(values) == self.ns:
                    params[name] = values
                else:
                    raise Exception("Not able to convert to pandas")
        for p1, i1 in self.pin_dic.items():
            for p2, i2 in self.pin_dic.items():
                params[(p1, p2)] = self.S[:, i1, i2]
        return pd.DataFrame(params)

    def _build_metadata(
        self,
        parameter_name_mapping: Optional[Dict[str, str]] = None,
        units: Optional[Dict[str, str]] = None,
    ) -> Dict:
        """Builds the metadata dict for the InPulse scattering matrix export

        Args:
            parameter_name_mapping (dict): mapping between parameter name in the model and exported file.
                The for of the dictionary is {<model-parameter-name>:<exported-parameter-name>}.
                Not mapped parameter are exported with their origina name.
            units (dict): Units of the exported parameters, in the form: {<exported-parameter-name>:<unit>}
                For parameters with no units explicit map to None is advised.

        Returns:
            dict: nested dictionary of the metadata
        """
        metadata = {
            "_schema": "InPulse S-Matrix RAW DATA",
            "_schema_version": "1.0",
            "_date": datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
            "_description": "Scattering matrix model exported from gensol",
            "units": {
                "abs2": None,
                "phase": "rad",
            },
        }
        parameter_name_mapping = (
            {} if parameter_name_mapping is None else parameter_name_mapping
        )
        units = {} if units is None else units
        solved_parameters_units = {
            parameter_name_mapping.get(par, par): None for par in self.solved_params
        }
        for par in solved_parameters_units:
            if par not in units:
                logger.warning(
                    f"Export to InPulse SM: parameter '{par}' has no unit. If None unit is correct, provide it explicitly."
                )
            else:
                solved_parameters_units[par] = units[par]
        metadata["units"].update(solved_parameters_units)
        metadata["port_modes"] = {
            name: [_[1] for _ in self.get_pin_modes(name)]
            for name in self.get_pin_basenames()
        }

        metadata["smatrix_map"] = {
            pinin: {
                pinout: {
                    modein[1]: {
                        modeout[1]: f"{pinin}_{modein[1]}//{pinout}_{modeout[1]}"
                        for modeout in self.get_pin_modes(pinout)
                    }
                    for modein in self.get_pin_modes(pinin)
                }
                for pinout in self.get_pin_basenames()
            }
            for pinin in self.get_pin_basenames()
        }

        return metadata

    def _build_data(self) -> pd.DataFrame:
        """Builds the dataframe used to export data to InPulse scattering matrix"""
        full_data = self.get_full_data()
        for column in full_data.columns:
            if column in self.solved_params:
                continue
            col = full_data.pop(column)
            fullin, fullout = column
            fullin = fullin.split("_")
            fullout = fullout.split("_")
            (pinin, modein) = (fullin[0], "") if len(fullin) == 1 else fullin
            (pinout, modeout) = (fullout[0], "") if len(fullout) == 1 else fullout
            name = f"{pinin}_{modein}//{pinout}_{modeout}"
            full_data[f"{name}:abs2"] = np.abs(col) ** 2.0
            full_data[f"{name}:phase"] = np.angle(col)
        return full_data

    def export_InPulse(
        self,
        filename: str = "exported_model.csvy",
        parameter_name_mapping: Optional[Dict[str, str]] = None,
        units: Optional[Dict[str, str]] = None,
    ) -> str:
        """Export scattering matrix in InPulse format

        Args:
            filename (str): name of the file to export. Default to exported_model.csvy
            parameter_name_mapping (dict): mapping between parameter name in the model and exported file.
                The for of the dictionary is {<model-parameter-name>:<exported-parameter-name>}.
                Not mapped parameter are exported with their origina name.
            units (dict): Units of the exported parameters, in the form: {<exported-parameter-name>:<unit>}
                For parameters with no units explicit map to None is advised.

        Returs:
            str: string representing the scattering matrix in InPulse
        """
        metadata = self._build_metadata(
            units=units, parameter_name_mapping=parameter_name_mapping
        )
        metadata = yaml.dump(metadata)

        data = self._build_data()
        if parameter_name_mapping is not None:
            data.rename(columns=parameter_name_mapping, inplace=True)
        data = data.to_csv(index=False)

        to_export = f"{metadata}---\n{data}"
        with open(filename, "w") as f:
            f.write(to_export)
        return to_export

    def get_monitor(
        self, input_dic: Dict[str, float], power: bool = True
    ) -> pd.DataFrame:
        """Function for returning data from monitors

        This function returns the mode coefficients if inputs and outputs of every structure selected
            as monitors

        Args:
            input_dic (dict): Dictionary of the input amplitudes. Format is
                {'pinname' : amplitude (float or complex)}. Missing pins are assumed to have 0 amplitude.

        Returns:
            pandas DataFrame: DataFrame with the amplitude at the ports. It has one column for each
                parameter given to solve plus two columns for monitor port.
        """
        params = {}
        if self.ns == 1:
            params = deepcopy(self.solved_params)
        else:
            for name, values in self.solved_params.items():
                if len(values) == 1:
                    params[name] = np.array([values[0] for i in range(self.ns)])
                elif len(values) == self.ns:
                    params[name] = values
                else:
                    raise Exception("Not able to convert to pandas")

        u, d = self.int_func(input_dic)
        for pin, i in self.monitor_mapping.items():
            params[f"{pin}_i"] = np.abs(u[:, i]) ** 2.0 if power else u[:, i]
            params[f"{pin}_o"] = np.abs(d[:, i]) ** 2.0 if power else d[:, i]

        pan = pd.DataFrame.from_dict(params)
        return pan

        return u, d


class Waveguide(Model):
    """Model of a simple waveguide"""

    def __init__(self, L: float, n: float = 1.0, wl: float = 1.0) -> None:
        """Initialize the model

        Args:
            L (float) : length of the waveguide
            n (float or complex): effective index of the waveguide
            wl (float) : default wavelength of the waveguide
        """
        self.pin_dic = {"a0": 0, "b0": 1}
        self.N = 2
        self.S = np.identity(self.N, complex)
        self.L = L
        self.param_dic = {"wl": wl}
        self.n = n
        self.default_params = deepcopy(self.param_dic)

    def create_S(self) -> np.ndarray:
        """Function for returning the scattering matrix of the model
        Returns:
            ndarray: Scattering matrix of the model
        """
        wl = self.param_dic["wl"]
        n = self.n
        self.S = np.zeros((self.N, self.N), complex)
        self.S[0, 1] = np.exp(2.0j * np.pi * n / wl * self.L)
        self.S[1, 0] = np.exp(2.0j * np.pi * n / wl * self.L)
        return self.S

    def __str__(self):
        """Formatter function for printing"""
        return f"Model of waveguide of lenght {self.L:.3f} and index {self.n:.3f} (id={id(self)})"


class UserWaveguide(Model):
    """Template for a user defined waveguide"""

    def __init__(
        self,
        L: float,
        func: Callable,
        param_dic: Dict[str, Any] = None,
        allowedmodes: Dict[str, Dict] = None,
    ) -> None:
        """Initialize the model

        Args:
            L (float): length of the waveguide
            func (function): index function of the waveguide
            param_dic (dict): dictionary of the default parameters to be used (common to all modes)
            allowedmodes (dict): Dict of allowed modes and settings. Form is name:extra.
                extra is a dictionary containing the extra parameters to be passed to func
                Default is for 1 mode, with no name and no parameters
        """

        self.allowed = {"": {}} if allowedmodes is None else allowedmodes
        self.pin_dic = {}

        for i, mode in enumerate(self.allowed):
            if mode == "":
                self.pin_dic["a0"] = 2 * i
                self.pin_dic["b0"] = 2 * i + 1
            else:
                self.pin_dic[f"a0_{mode}"] = 2 * i
                self.pin_dic[f"b0_{mode}"] = 2 * i + 1

        self.N = len(self.pin_dic)
        self.S = np.identity(self.N, complex)
        self.param_dic = deepcopy(param_dic) if param_dic else {}
        self.default_params = deepcopy(self.param_dic)

        self.index_func = func
        self.L = L

    def create_S(self) -> np.ndarray:
        """Created the scattering Matrix"""
        wl = self.param_dic["wl"]
        S_list = []
        for mode, extra in self.allowed.items():
            self.param_dic.update(extra)
            n = self.index_func(**self.param_dic)
            S = np.zeros((2, 2), complex)
            S[0, 1] = np.exp(2.0j * np.pi * n / wl * self.L)
            S[1, 0] = np.exp(2.0j * np.pi * n / wl * self.L)
            S_list.append(S)
        self.S = diag_blocks(S_list)
        return self.S


class BeamSplitter(Model):
    """Model of variable ration beam splitter"""

    def __init__(self, ratio: float = 0.5, t: float = None, phase: float = 0.0) -> None:
        """Initialize the model

        Args:
            ratio (float) : Power coupling coefficient. It is also the splitting ratio if t is not provided.
            t (float): Power transmission coefficient. If None (default) it is calculated from the ratio assuming
                no loss in the component.
            phase (float) : phase shift of the transmitted ray (in unit of pi). Default to 0.0
        """
        self.pin_dic = {"a0": 0, "a1": 1, "b0": 2, "b1": 3}
        self.N = 4
        self.ratio = ratio
        self.phase = phase
        p1 = np.pi * self.phase
        c = 1.0j * np.sqrt(self.ratio)
        t = np.sqrt(1.0 - self.ratio) if t is None else np.sqrt(t)
        self.S = np.zeros((self.N, self.N), complex)
        self.param_dic = {}
        self.default_params = deepcopy(self.param_dic)
        # self.S[:2,2:]=np.array([[t,c],[c,-t]])
        # self.S[2:,:2]=np.array([[t,c],[c,-t]])
        self.S[:2, 2:] = np.exp(2.0j * np.pi * phase) * np.array([[t, c], [c, t]])
        self.S[2:, :2] = np.exp(2.0j * np.pi * phase) * np.array([[t, c], [c, t]])

    def __str__(self):
        """Formatter function for printing"""
        return f"Model of beam-splitter with ratio {self.ratio:.3} (id={id(self)})"


class Splitter1x2(Model):
    """Model of 1x2 Splitter"""

    def __init__(self) -> None:
        """Initialize the model"""
        self.pin_dic = {"a0": 0, "b0": 1, "b1": 2}
        self.N = 3
        self.S = (
            1.0
            / np.sqrt(2.0)
            * np.array([[0.0, 1.0, 1.0], [1.0, 0.0, 0.0], [1.0, 0.0, 0.0]], complex)
        )
        self.param_dic = {}
        self.default_params = deepcopy(self.param_dic)

    def __str__(self):
        return f"Model of 1x2 splitter (id={id(self)})"


class Splitter1x2Gen(Model):
    """Model of 1x2 Splitter with possible reflection between the 2 port side.
    TODO: verify this model makes sense
    """

    def __init__(self, cross: float = 0.0, phase: float = 0.0) -> None:
        """Initialize the model

        Args:
            cross (float) : ratio of reflection (power ratio)
            phase (float) : phase shift of the reflected ray (in unit of pi)
        """
        self.pin_dic = {"a0": 0, "b0": 1, "b1": 2}
        self.N = 3
        self.param_dic = {}
        self.default_params = deepcopy(self.param_dic)
        c = np.sqrt(cross)
        t = np.sqrt(0.5 - cross)
        p1 = np.pi * phase
        self.S = np.array(
            [
                [0.0, t, t],
                [t, 0.0, c * np.exp(1.0j * p1)],
                [t, c * np.exp(-1.0j * p1), 0.0],
            ],
            complex,
        )


class PhaseShifter(Model):
    """Model of multimode variable phase shifter"""

    def __init__(self, param_name: str = "PS", param_default: float = 0.0) -> None:
        """Initialize the model

        Args:
            param_name (str): name of the parameter of the Phase Shifter
            param_default (float): default value of the Phase Shift in pi units
        """
        self.param_dic = {}
        self.pin_dic = {"a0": 0, "b0": 1}
        self.N = 2
        self.pn = param_name
        self.param_dic = {param_name: param_default}
        self.default_params = deepcopy(self.param_dic)

    def create_S(self) -> np.ndarray:
        """Function for returning the scattering matrix of the model
        Returns:
            ndarray: Scattering matrix of the model
        """
        S = np.zeros((self.N, self.N), complex)
        S[0, 1] = np.exp(1.0j * np.pi * self.param_dic[self.pn])
        S[1, 0] = np.exp(1.0j * np.pi * self.param_dic[self.pn])
        self.S = S
        return self.S


class PushPullPhaseShifter(Model):
    """Model of multimode variable phase shifter"""

    def __init__(self, param_name: str = "PS") -> None:
        """Initialize the model

        Args:
            param_name (str) : name of the parameter of the Phase Shifter
        """
        self.param_dic = {}
        self.pin_dic = {"a0": 0, "b0": 1, "a1": 2, "b1": 3}
        self.N = 4
        self.pn = param_name
        self.param_dic = {param_name: 0.0}
        self.default_params = deepcopy(self.param_dic)

    def create_S(self) -> np.ndarray:
        """Function for returning the scattering matrix of the model
        Returns:
            ndarray: Scattering matrix of the model
        """
        S1 = np.zeros((2, 2), complex)
        S1[0, 1] = np.exp(0.5j * np.pi * self.param_dic[self.pn])
        S1[1, 0] = np.exp(0.5j * np.pi * self.param_dic[self.pn])
        S2 = np.zeros((2, 2), complex)
        S2[0, 1] = np.exp(-0.5j * np.pi * self.param_dic[self.pn])
        S2[1, 0] = np.exp(-0.5j * np.pi * self.param_dic[self.pn])
        self.S = diag_blocks([S1, S2])
        return self.S

    def __str__(self):
        """Formatter function for printing"""
        return f"Model of variable Push-Pull phase shifter (id={id(self)})"


class PolRot(Model):
    """Model of a 2 modes polarization rotator"""

    def __init__(self, angle: float = None, angle_name: str = "angle") -> None:
        """Initialize the model

        If angle is provided the rotation is fixed to that value. If not, the rotation is assumed
            variable and the angle will be fetched form the parameter dictionary.

        Args:
            angle (float) : fixed value of the rotation angle (in pi units). Default is None
            angle_name (str) : name of the angle parameter
        """
        self.pin_dic = {"a0_pol0": 0, "a0_pol1": 1, "b0_pol0": 2, "b0_pol1": 3}
        self.N = 4
        self.param_dic = {}
        if angle is None:
            self.fixed = False
            self.angle_name = angle_name
            self.param_dic = {angle_name: 0.0}
        else:
            self.fixed = True
            c = np.cos(np.pi * angle)
            s = np.sin(np.pi * angle)
            self.S = np.zeros((self.N, self.N), complex)
            self.S[:2, 2:] = np.array([[c, s], [-s, c]])
            self.S[2:, :2] = np.array([[c, -s], [s, c]])
        self.default_params = deepcopy(self.param_dic)

    def create_S(self) -> np.ndarray:
        """Function for returning the scattering matrix of the model

        Returns:
            ndarray: Scattering matrix of the model
        """
        if self.fixed:
            return self.S
        else:
            angle = self.param_dic[self.angle_name]
            c = np.cos(np.pi * angle)
            s = np.sin(np.pi * angle)
            S = np.zeros((self.N, self.N), complex)
            S[:2, 2:] = np.array([[c, s], [-s, c]])
            S[2:, :2] = np.array([[c, -s], [s, c]])
            return S


class Attenuator(Model):
    """Model of attenuator in dB"""

    def __init__(self, loss: float = 0.0) -> None:
        """Initialize the model

        Args:
            loss: value of the loss (in dB)
        """
        self.param_dic = {}
        self.pin_dic = {"a0": 0, "b0": 1}
        self.N = 2
        self.loss = loss
        self.S = np.zeros((self.N, self.N), complex)
        self.S[0, 1] = 10.0 ** (-0.05 * loss)
        self.S[1, 0] = 10.0 ** (-0.05 * loss)
        self.default_params = deepcopy(self.param_dic)


class LinearAttenuator(Model):
    """Model of attenuator in absolute unit"""

    def __init__(self, c: float = 1.0) -> None:
        """Initialize the model

        Args:
            c (float): fraction of power transmitted:
                1.0 -> no loss
                0.3 -> 30% of the power is transmitted
                0.0 -> no light transmitted
        """
        self.param_dic = {}
        self.pin_dic = {"a0": 0, "b0": 1}
        self.N = 2
        self.S = np.zeros((self.N, self.N), complex)
        self.S[0, 1] = np.sqrt(c)
        self.S[1, 0] = np.sqrt(c)
        self.default_params = deepcopy(self.param_dic)


class Mirror(Model):
    """Model of partially reflected Mirror"""

    def __init__(self, ref: float = 0.5, phase: float = 0.0) -> None:
        """Initialize the model

        Args:
            ref (float) : ratio of reflected power
            phase (float): phase shift of the reflected ray (in pi units)
        """
        self.pin_dic = {"a0": 0, "b0": 1}
        self.param_dic = {}
        self.default_params = deepcopy(self.param_dic)
        self.N = 2
        self.ref = ref
        self.phase = phase
        t = np.sqrt(self.ref)
        c = np.sqrt(1.0 - self.ref)
        p1 = np.pi * self.phase
        self.S = np.array(
            [[t * np.exp(1.0j * p1), c], [-c, t * np.exp(-1.0j * p1)]], complex
        )


class PerfectMirror(Model):
    """Model of perfect mirror (only one port), 100% reflection"""

    def __init__(self, phase: float = 0.0) -> None:
        """Initialize the model

        Args:
            phase (float): phase of the reflected ray (in pi unit)
        """
        self.pin_dic = {"a0": 0}
        self.param_dic = {}
        self.default_params = deepcopy(self.param_dic)
        self.N = 1
        self.phase = phase
        p1 = np.pi * self.phase
        self.S = np.array([[np.exp(1.0j * p1)]], complex)


class FPR_NxM(Model):
    """Model of Free Propagation Region. TODO: check this model makes sense"""

    def __init__(self, N: int, M: int, phi: float = 0.1) -> None:
        """Initialize the model

        Args:
            N (int) : number of input ports
            M (int) : number of output ports
            phi (float) : phase difference between adjacent ports
        """
        self.param_dic = {}
        self.default_params = deepcopy(self.param_dic)
        self.pin_dic = {f"a{i}": i for i in range(N)}
        self.pin_dic.update({f"b{i}": N + i for i in range(M)})
        Sint = np.zeros((N, M), complex)
        for i in range(N):
            for j in range(M):
                Sint[i, j] = np.exp(
                    -1.0j * np.pi * phi * (i - 0.5 * N + 0.5) * (j - 0.5 * M + 0.5)
                )
        Sint2 = np.conj(np.transpose(Sint))
        self.S = np.concatenate(
            [
                np.concatenate([np.zeros((N, N), complex), Sint / np.sqrt(M)], axis=1),
                np.concatenate([Sint2 / np.sqrt(N), np.zeros((M, M), complex)], axis=1),
            ],
            axis=0,
        )


class Ring(Model):
    """Model of ring resonator filter"""

    def __init__(self, R: float, n: float, alpha: float, t: float) -> None:
        """Initialize the model

        Args:
            R (float) : radius of the ring
            n (float) : effective index of the waveguide in the ring
            alpha (float) : one trip loss coefficient (remaining complex amplitude)
            t (float) : transmission of the beam splitter (complex amplitude)
        """
        self.pin_dic = {"a0": 0, "b0": 1}
        self.N = 2
        self.S = np.identity(self.N, complex)
        self.R = R
        self.n = n
        self.alpha = alpha
        self.t = t
        self.param_dic = {"wl": None}
        self.default_params = deepcopy(self.param_dic)

    def create_S(self) -> np.ndarray:
        """Function for returning the scattering matrix of the model
        Returns:
            ndarray: Scattering matrix of the model
        """
        wl = self.param_dic["wl"]
        n = self.n
        t = self.t
        ex = np.exp(-4.0j * np.pi**2.0 / wl * n * self.R)
        b = (-self.alpha + t * ex) / (-self.alpha * t + ex)
        self.S = np.zeros((self.N, self.N), complex)
        self.S[0, 1] = b
        self.S[1, 0] = b
        return self.S


class TH_PhaseShifter(Model):
    """Model of thermal phase shifter (dispersive waveguide + phase shifter)"""

    def __init__(
        self,
        L: float,
        Neff: Callable,
        R: float = None,
        w: float = None,
        wl: float = None,
        pol: int = None,
        param_name: str = "PS",
    ) -> None:
        """Initialize the model

        Args:
            L (float) : length of the waveguide
            Neff (function): function returning the effective index of the wavegude. It must be a function of
                wl,R,w, and pol
            wl (float) : default wavelength of the waveguide
            w  (float) : default width of the waveguide
            R  (float) : default bending radius of the waveguide
            pol (int)  : default mode of the waveguide
            param_name (str) : name of the parameter of the Phase Shifter
        """
        self.pin_dic = {"a0": 0, "b0": 1}
        self.N = 2
        self.Neff = Neff
        self.L = L
        self.pn = param_name
        self.param_dic = {"R": R, "w": w, "wl": wl, "pol": pol, param_name: 0.0}
        self.default_params = deepcopy(self.param_dic)

    def create_S(self) -> np.ndarray:
        """Function for returning the scattering matrix of the model
        Returns:
            ndarray: Scattering matrix of the model
        """
        wl = self.param_dic["wl"]
        n = self.Neff(**self.param_dic)
        self.S = np.zeros((self.N, self.N), complex)
        self.S[0, 1] = np.exp(
            1.0j * np.pi * (2.0 * n / wl * self.L + self.param_dic[self.pn])
        )
        self.S[1, 0] = np.exp(
            1.0j * np.pi * (2.0 * n / wl * self.L + self.param_dic[self.pn])
        )
        return self.S


class AWGfromVPI(Model):
    """Define custom model from a VPI scattering matrix format"""

    def __init__(self, filename: str, pol: str = "TE", fsr: float = 1.0) -> None:
        """Instantiate an AWG from a VPI S-matrix.

        Args:
            filename (str): Name of the file where the S-matrix is stored.
            pol (str): Polarization that is considered, while the other value is discarded. Default is TE.
            fsr (float): ratio between the provided portion spectrum and the fpr of the AWG.
                if wavelength outside the spectrum are provided they will be reconstructed assuming periodicity.
        """
        with open(filename) as f:
            data = f.read()

            data = data.split("\n\n")[2:]

        coeff = {}
        pins = []
        for t in data:
            if "NullTransferFunction" in t:
                continue
            p = t.split("\n")
            if len(p) == 1:
                continue
            pin_in = p[0]
            pin_out = p[1]
            if pol not in pin_in:
                continue
            if pol not in pin_out:
                continue
            pin_in = (
                pin_in.split()[2].split("#")[0][0] + pin_in.split()[2].split("#")[1]
            )
            pin_out = (
                pin_out.split()[2].split("#")[0][0] + pin_out.split()[2].split("#")[-1]
            )
            if pin_in not in pins:
                pins.append(pin_in)
            if pin_out not in pins:
                pins.append(pin_out)
            dd = io.StringIO(t)
            ar = np.loadtxt(dd)
            LAM = ar[:, 0]
            coeff[(pin_in, pin_out)] = ar[:, 1] * np.exp(
                1.0j * np.pi / 180 * 0 * ar[:, 2]
            )

        pins.sort()
        S = np.zeros((len(LAM), len(pins), len(pins)), dtype=complex)
        for i, ipin in enumerate(pins):
            for j, jpin in enumerate(pins):
                if (ipin, jpin) in coeff:
                    S[:, i, j] = coeff[(ipin, jpin)]

        self.pin_dic = {pin: i for i, pin in enumerate(pins)}
        self.N = len(pins)
        self.param_dic = {}
        self.default_params = {}

        self.S_func = interp1d(LAM, S, axis=0)

        self.fsr = (LAM[-1] - LAM[0]) / fsr
        self.lamc = (LAM[-1] + LAM[0]) / 2.0

    def create_S(self) -> np.ndarray:
        lam = self.param_dic["wl"]
        self.S = self.S_func(
            self.lamc
            - 0.5 * self.fsr
            + np.mod(lam - self.lamc + 0.5 * self.fsr, self.fsr)
        )
        return self.S


class FPR(Model):
    """Simplified model of FPR circle mount"""

    def __init__(
        self, n: int, m: int, R: float, d1: float, d2: float, Ri: float = None
    ) -> None:
        """Initialize the model

        Args:
            n (int): Number of inputs arms.
            m (int): Number of output arms.
            R (float): Radius the output part of FPR (input and output are on the same radius)
            d1 (float): Distance of inputs on the input plane.
            d2 (float): Distance of output on the output plane.

        Returns:
            Model: Model of the FPR region

        """
        Ri = 0.5 * R if Ri is None else Ri
        self.n, self.m = n, m
        self.d1, self.d2 = d1, d2
        self.R = R
        d1 = {f"a{i:03}": i for i in range(n)}
        d2 = {f"b{i:03}": n + i for i in range(m)}
        dic = {**d1, **d2}
        super().__init__(pin_dic=dic)
        t1 = self.d1 / Ri * np.array(line(n))
        t2 = self.d2 / R * np.array(line(m))
        T1, T2 = np.meshgrid(t1, t2, indexing="ij")
        DY = R * np.sin(T2) - Ri * np.sin(T1)
        DX = R * np.cos(T2) - Ri * (1.0 - np.cos(T1))
        self.DR = np.sqrt(DY**2.0 + DX**2.0)
        self.S = np.zeros((n + m, n + m), dtype=complex)

    def create_S(self) -> np.ndarray:
        """Creates the scattering matrix

        Returns:
            2darray: Scattering matrix of the FPR

        """
        lam = self.param_dic["wl"]
        n = self.n
        Mat = np.exp(2.0j * np.pi / lam * self.DR) / np.sqrt(max(self.m, self.n))
        self.S[:n, n:] = Mat
        self.S[n:, :n] = np.transpose(Mat)
        return self.S


class FPRGaussian(Model):
    """Simplified model of FPR circle mount based on Gaussian beams."""

    def __init__(
        self,
        n: int,
        m: int,
        R: float,
        d1: float,
        d2: float,
        w1: float,
        w2: float,
        n_slab: float,
        Ri: float = None,
    ) -> None:
        """Initialize the model

        Args:
            n (int): Number of inputs arms.
            m (int): Number of output arms.
            R (float): Radius the output part of FPR (input and output are on the same radius).
            d1 (float): Center-to-center distance of inputs on the input plane.
            d2 (float): Center-to-center distance of output on the output plane in.
            n_slab (float | callable): Effective index of the slab mode.
            R_i (float): Input radius. If it is not specified it is set equal to the output one. Default is None.

        Returns:
            Model: Model of the FPR region

        """
        Ri = R if Ri is None else Ri
        self.n, self.m = n, m
        self.d1, self.d2 = d1, d2
        self.Ri = Ri
        self.R = R
        self.n_slab = n_slab
        self.w1 = w1
        self.w2 = w2
        d1 = {f"a{i:03}": i for i in range(n)}
        d2 = {f"b{i:03}": n + i for i in range(m)}
        dic = {**d1, **d2}
        super().__init__(pin_dic=dic)

        self.t1 = self.d1 / Ri * line(n)
        self.t2 = self.d2 / R * line(m)
        self.pos1 = [
            (Ri * (1 - np.cos(self.t1[i])), Ri * np.sin(self.t1[i])) for i in range(n)
        ]
        self.pos2 = [(R * np.cos(self.t2[i]), R * np.sin(self.t2[i])) for i in range(m)]

        self.S = np.zeros((n + m, n + m), dtype=complex)

    def create_S(self):
        """Creates the scattering matrix

        Returns:
            2darray: Scattering matrix of the FPR

        """
        lam = self.param_dic["wl"]
        if type(self.n_slab) == float or type(self.n_slab) == int:
            n_slab = self.n_slab
        elif callable(self.n_slab):
            n_slab = self.n_slab(lam)

        n = self.n
        m = self.m
        mat = np.zeros((n, m), dtype=complex)
        z = 0.5 * self.R
        for i in range(n):
            beam1 = GaussianBeam(
                w0=self.w1,
                n=n_slab,
                wl=lam,
                z0=self.pos1[i][0],
                x0=self.pos1[i][1],
                theta=-np.rad2deg(self.t1[i]),
            )
            for j in range(m):
                beam2 = GaussianBeam(
                    w0=self.w2,
                    n=n_slab,
                    wl=lam,
                    z0=self.pos2[j][0],
                    x0=self.pos2[j][1],
                    theta=np.rad2deg(self.t2[j]),
                )

                def to_integrate(x):
                    integrand = beam1.field(z=z, x=x) * np.conjugate(
                        beam2.field(z=z, x=x)
                    )
                    return np.array([integrand.real, integrand.imag])

                res = quad_vec(to_integrate, -np.inf, np.inf)
                mat[i, j] = res[0][0] + 1j * res[0][1]

        self.S[:n, n:] = mat
        self.S[n:, :n] = np.transpose(mat)
        return self.S

    def show(self, ax: matplotlib.axes.Axes = None) -> matplotlib.axes.Axes:
        """Shows the input and output ports of the FPR.

        Args:
            ax (matplotlib.axes.Axes): Axis onto which the plot is mapped.

        Returns:
            matplotlib.axes.Axes: The axis.

        """
        x_in = [el[0] for el in self.pos1]
        x_out = [el[0] for el in self.pos2]
        y_in = [el[1] for el in self.pos1]
        y_out = [el[1] for el in self.pos2]
        if ax in None:
            fig, ax = plt.subplots()
        ax.scatter(x_in, y_in, label="input")
        ax.scatter(x_out, y_out, label="output")
        ax.legend(loc="lower center")
        ax.set_xlabel("x [$\mu$m]")
        ax.set_ylabel("y [$\mu$m]")
        ax.set_title(f"FPR: R$_i$={self.Ri} $\mu$m, R$_o$={self.R} $\mu$m")

        return ax


class CWA(Model):
    """Simple model fo a Coupled Waveguide Array"""

    def __init__(self, N: int, L: float, n: float = 1.0, k: float = 0.1):
        """Initialize the model

        Args:
            N (int): Number of waveguides in the array.
            L (float): Length of the CWA section
            n (float): Index of the single waveguide in the array.
            k (float): Coupling coefficient between waveguides.

        Returns:
            Model: Model of a CWA

        """
        self.NW = N
        self.n = n
        self.k = k
        self.L = L
        d1 = {f"a{i}": i for i in range(N)}
        d2 = {f"b{i}": N + i for i in range(N)}
        dic = {**d1, **d2}
        super().__init__(pin_dic=dic)
        i, j = list(range(N)), list(range(N))
        I, J = np.meshgrid(i, j, indexing="ij")
        self.RM = np.exp(2.0j * np.pi * I * J / self.NW) / np.sqrt(self.NW)
        self.S = np.zeros((2 * N, 2 * N), dtype=complex)

    def create_S(self) -> np.ndarray:
        B = [
            2.0 * np.pi * self.n / self.param_dic.get("wl", 1.0)
            + 2.0 * self.k * np.cos(2.0 * np.pi * n / self.NW)
            for n in range(self.NW)
        ]
        P = np.diag(np.exp(1.0j * np.array(B) * self.L))
        S = np.dot(np.dot(self.RM, P), self.RM)
        self.S[: self.NW, self.NW :] = S
        self.S[self.NW :, : self.NW] = S
        return self.S


class Model_from_InPulse(Model):
    """Class for model from InPulse S-Matrix RAW DATA file"""

    def __init__(
        self,
        file: str,
        parameter_name_mapping: Optional[Dict[str, str]] = None,
        mode_mapping: Optional[Dict[str, str]] = None,
    ) -> None:
        """
        Initialize the model

        Args:
            file (str): Path of the file to be loaded.
            parameter_name_mapping (dict[str,str]): mapping of the parameter name between the file and the model.
                The format is {'<file-parameter-name>':'<model-parameter-name>'}
                Parameters not provided retain theur original name.
                Example:
                Wavelength is usually called 'wl' in gensol, but it may be 'wavelength' in the file.
                To uniform the model to the other provide the mapping {"wavelength":"wl"}
            mode_mapping (dict[str,str]): mapping between the mode names in the file and the model.
                The format is {'<file-mode-name>':'<model-mode-name>'}
                If nothing is provided, all the modes are loaded with their original name.
                This will create multiple pins in the model with name '<pin-name>_<mode-name>'
                If a dictionary is provided, only the mapped modes are loaded and the names are changed accordingly.
                If a mode is mapped to the empty string ("" or ''), the result port name is just <pin-name>

        Returns:
            None
        """
        self._file = file
        parameter_name_mapping = (
            {} if parameter_name_mapping is None else parameter_name_mapping
        )
        pin_dic, param_dic = self._process_file()
        self.parameters_names = [
            parameter_name_mapping.get(par, par) for par in self.parameters_names
        ]
        if mode_mapping is not None:
            pin_mapping = {}
            for pin in pin_dic:
                try:
                    name, mode = pin.split("_")
                except ValueError:
                    name, mode = pin, ""
                mode = mode_mapping.get(mode)
                if mode is not None:
                    pin_mapping[pin] = name if mode == "" else f"{name}_{mode}"

            pin_dic = {target: i for i, (pin, target) in enumerate(pin_mapping.items())}

            new_map_columns = {}
            for (pin_in, pin_out), column in self.map_columns.items():
                new_in = pin_mapping.get(pin_in)
                new_out = pin_mapping.get(pin_out)
                if new_in is not None and new_out is not None:
                    new_map_columns[(new_in, new_out)] = column
            self.map_columns = new_map_columns

        if not pin_dic:
            raise ValueError(
                "Loaded model from InPulse has no pins. Hint: check mode mapping."
            )

        super().__init__(pin_dic=pin_dic, param_dic=param_dic)

    def __get_pin_dic(self, port_modes: Dict) -> Dict[str, int]:
        """
        Creates the pin dictionary of the model from the file

        Args:
            port_modes (dict): The nested dictionary containing the "port_modes" part of the metadata

        Returns:
            dict: pin dictionary in the form {<pin-name> : <pin-number>}
        """
        pin_dic = {}
        _i = 0
        for pin, modes in port_modes.items():
            for mode in modes:
                name = pin if mode == "" else f"{pin}_{mode}"
                pin_dic[name] = _i
                _i += 1
        return pin_dic

    def __get_pin_map(
        self, smatrix_map: Dict
    ) -> Tuple[Set[str], Dict[Tuple[str, str], str]]:
        """
        Retrieve the mapping between the ports and the columns in the data section.

        Args:
            smatrix_map (dict): The nested dictionary containing the "smatrix_map" part of the metadata

        Returns:
            set: set of the columns names
            map_columns: map between pairs of pins and the related column
        """
        map_columns = {}
        columns = set()
        for pin, target in smatrix_map.items():
            for pin_target, modes in target.items():
                for modein, modes_out in modes.items():
                    for modeout, column in modes_out.items():
                        if column:
                            name1 = pin if modein == "" else f"{pin}_{modein}"
                            name2 = (
                                pin_target
                                if modeout == ""
                                else f"{pin_target}_{modeout}"
                            )
                            map_columns[(name1, name2)] = column
                            columns.add(column)
        return columns, map_columns

    def _process_file(self) -> Tuple[Dict, Dict]:
        """
        Load the data from the file and proccess the content

        Args;
            None

        Returns:
            dict: pin dictionry of the model
            dict: parameters dictionary of the models
        """
        with open(self._file, "r") as file:
            text = file.read()
        metadata, data = text.split("---")
        metadata = metadata.replace("\t", "    ")
        metadata = yaml.safe_load(metadata)

        pin_dic = self.__get_pin_dic(metadata["port_modes"])
        columns, self.map_columns = self.__get_pin_map(metadata["smatrix_map"])

        param_dic = {
            name: parameter_info["mean"]
            for name, parameter_info in metadata.get("csm_parameters", {}).items()
        }

        data = io.StringIO(data)
        data = pd.read_csv(data)
        self.parameters_names = [col for col in data.columns if ":" not in col]
        parameters_values = data[self.parameters_names].values
        self.functions_columns = {}
        for column in columns:
            ampl = np.asarray(data[f"{column}:abs2"].values)
            phase = np.asarray(data[f"{column}:phase"].values)
            amplitude = np.sqrt(ampl) * np.exp(1.0j * phase)
            if len(self.parameters_names) >= 2:
                self.functions_columns[column] = LinearNDInterpolator(
                    parameters_values, amplitude
                )
            else:
                self.functions_columns[column] = interp1d(
                    np.squeeze(parameters_values), amplitude
                )

        return pin_dic, param_dic

    def create_S(self) -> np.ndarray:
        """Returns the scattering matrix of the model"""
        S = np.zeros((self.N, self.N), dtype=complex)
        parameters_values = tuple(
            self.param_dic[parameter] for parameter in self.parameters_names
        )
        amplitudes = {
            col: func(*parameters_values)
            for col, func in self.functions_columns.items()
        }

        for (pin_in, pin_out), column in self.map_columns.items():
            S[self.pin_dic[pin_in], self.pin_dic[pin_out]] = amplitudes[column]
        self.S = S
        return S



# -------------------------------------------
#
# GenSol Package
#
# Python tool for simulation of abstract photonic circuits
#
# @author: Marco Passoni
#
# -------------------------------------------


""" Initialization file
"""
import os

try:
    from ._version import __version__
except ModuleNotFoundError:
    from hatch_vcs.version_source import VCSVersionSource
    import toml

    basedir = os.path.split(os.path.dirname(__file__))[0]
    config = toml.load(os.path.join(basedir, "pyproject.toml"))
    vcs_version = VCSVersionSource(basedir, config["tool"]["hatch"]["version"])
    __version__ = vcs_version.get_version_data()["version"]

import logging

logger = logging.getLogger(__name__)

sol_list = []

from .log import logfile, debugfile
from .scattering import S_matrix
from .structure import Structure
from .sol import *
from .model import *



import logging
from lekkersim import logger
import __main__
import lekkersim as lk

# cleaning logger. Needed if run within Spyder for some reason
logger.handlers.clear()
logger.filters = []

logger.setLevel(logging.DEBUG)

std = logging.StreamHandler()
std.setLevel(logging.INFO)
fmt = logging.Formatter("%(levelname)s:%(name)s:%(message)s")
std.setFormatter(fmt)


logger.addHandler(std)
logger.info(f"version {lk.__version__}")


class DuplicateFilter:
    """Filter for removing double entry in logger. Works on WARNING level or higher"""

    def __init__(self):
        self.msgs = set()

    def filter(self, record):
        """Filtering function"""
        if record.levelno < 30:
            return True
        rv = record.msg not in self.msgs
        self.msgs.add(record.msg)
        return rv


dup_filter = DuplicateFilter()
logger.addFilter(dup_filter)


def log_filter_on():
    """Activate filter to remove double log entries (WARNING and above)"""
    logger.addFilter(dup_filter)


def log_filter_off():
    """Deactivate filter to remove double log entries"""
    logger.removeFilter(dup_filter)


def logfile(filename=None, stdout=False, level=logging.INFO):
    """Switch log output to file

    Args:
        filename (str): name of the file. Default is the name of the running python file + '.log'
        std (bool): if True, output to stdout is also kept. Default is false
        level (int): level of the file handler, default is logging.WARNING (30)
    """
    filename = f"{__main__.__file__}.log" if filename is None else filename
    han = logging.FileHandler(filename, mode="w")
    han.setFormatter(fmt)
    logger.addHandler(han)
    han.setLevel(level)
    logger.removeHandler(std)
    log_filter_off()
    logger.info(f"version {lk.__version__}")
    log_filter_on()
    if stdout:
        logger.addHandler(std)


def debugfile(filename=None):
    """Set up and additional logfile for debug

    Args:
        filename (str): name of the file. Default is the name of the running python file + '.dbg'
    """
    filename = f"{__main__.__file__}.dbg" if filename is None else filename
    han = logging.FileHandler(filename, mode="w")
    han.setLevel(logging.DEBUG)
    han.setFormatter(fmt)
    logger.addHandler(han)
